import re
import uuid
from logging import Logger
from typing import Awaitable, Callable, Dict, TypeAlias, TypeVar
from uuid import uuid4

import json_logging
from connexion import FlaskApp
from flask import Flask, Request, Response, request
from flask.typing import ResponseReturnValue
from injector import inject

from ...config import Config
from ..consts import (
    CONTENT_SECURITY_POLICY_HEADER,
    CORRELATION_ID_HEADER,
    CORS_ACCESS_CONTROL_ALLOW_CREDENTIALS_HEADER,
    CORS_ACCESS_CONTROL_ALLOW_METHODS_HEADER,
    CORS_ACCESS_CONTROL_ALLOW_ORIGIN_HEADER,
    HOST_HEADER,
    INCOMING_REQUEST_MESSAGE,
    ORIGIN_HEADER,
    OUTGOING_RESPONSE_MESSAGE,
    REQUEST_COOKIE_HEADER,
    RESPONSE_COOKIE_HEADER,
)

# pyright: reportUnusedFunction=false


# Fixes type problems when using @inject with @app.before_request and @app.after_request.
# The main difference with these types as opposed to the Flask-defined types is that
# these types allow the handler to take any arguments, versus no arguments or just Response.
AfterRequestCallable: TypeAlias = (
    Callable[..., Response] | Callable[..., Awaitable[Response]]
)
BeforeRequestCallable: TypeAlias = (
    Callable[..., ResponseReturnValue | None]
    | Callable[..., Awaitable[ResponseReturnValue | None]]
)
T_request_callable = TypeVar(
    "T_request_callable", bound=BeforeRequestCallable | AfterRequestCallable | None
)
# Fixes type problems when using Flask-Injector, which automatically sets up any bound
# error handlers. The error handler types in Flask have the same problem as After/BeforeRequestCallable
# in that their parameters are `[Any]` rather than `...`. This relaxes that. The type
# is valid here because Flask-Injector does actually include the provided types as parameters.
ErrorHandlerCallable = (
    Callable[..., ResponseReturnValue] | Callable[..., Awaitable[ResponseReturnValue]]
)
T_error_handler = TypeVar("T_error_handler", bound=ErrorHandlerCallable)

TFlaskApp = Flask | FlaskApp
T_flask_app = TypeVar("T_flask_app", bound=TFlaskApp)


def bind_requesthandler(
    app: T_flask_app,
    flask_app_handler: Callable[
        [T_flask_app, Callable[..., Response | None]], T_request_callable
    ],
):
    def wrapper(request_callable: Callable[..., Response | None]) -> T_request_callable:
        return flask_app_handler(app, request_callable)

    return wrapper


def _get_correlation_id(log: Logger) -> str:
    correlation_id = _get_correlation_id_from_json_logging(log)

    if not correlation_id:
        correlation_id = _get_correlation_id_from_headers(log)

    return correlation_id


def _get_correlation_id_from_headers(log: Logger) -> str:
    try:
        correlation_id = request.headers.get(CORRELATION_ID_HEADER)

        if correlation_id:
            # validate format
            _ = uuid.UUID(correlation_id)
        else:
            correlation_id = str(uuid4())
            log.info(
                f'Generated new UUID "{correlation_id}" for {CORRELATION_ID_HEADER} request header.'
            )

        return correlation_id

    except ValueError as e:
        log.warning(f"Badly formatted {CORRELATION_ID_HEADER} received in request.")
        raise e


def _get_correlation_id_from_json_logging(log: Logger) -> str | None:
    correlation_id: None | str
    try:
        correlation_id = json_logging.get_correlation_id(request)
        # validate format
        _ = uuid.UUID(correlation_id)
        return correlation_id
    except ValueError as e:
        log.warning(f"Badly formatted {CORRELATION_ID_HEADER} received in request.")
        raise e
    except Exception as e:
        log.debug(
            f"Error received when getting {CORRELATION_ID_HEADER} header from `json_logging`. Possibly `json_logging` is not configured, and this is not an error.",
            exc_info=e,
        )


@inject
def _log_all_api_requests(
    request: Request,
    config: Config,
    log: Logger,
):
    request_headers_safe: dict[str, str] = dict(request.headers)

    correlation_id = _get_correlation_id(log)

    if (
        request_headers_safe.get(REQUEST_COOKIE_HEADER)
        and config.flask
        and config.flask.session
    ):
        request_headers_safe[REQUEST_COOKIE_HEADER] = re.sub(
            rf"({config.flask.session.cookie.name}=)[^;]+(;|$)",
            r"\1<redacted>\2",
            request_headers_safe[REQUEST_COOKIE_HEADER],
        )

    log.info(
        INCOMING_REQUEST_MESSAGE,
        request.method,
        request.url,
        request.host,
        request.remote_addr,
        request.remote_user,
        extra={
            "props": {
                "correlation_id": correlation_id,
                "headers": request_headers_safe,
            }
        },
    )


@inject
def _ordered_api_response_handers(response: Response, config: Config, log: Logger):
    _wrap_all_api_responses(response, config, log)
    _log_all_api_responses(response, config, log)
    return response


def _wrap_all_api_responses(response: Response, config: Config, log: Logger):
    correlation_id = _get_correlation_id(log)

    cors_domains: list[str] | None = None
    if config.web.security.cors.origins:
        cors_domains = config.web.security.cors.origins
    else:
        if not response.headers.get(CORS_ACCESS_CONTROL_ALLOW_ORIGIN_HEADER):
            cors_domains = (
                (header_value := request.headers.get(ORIGIN_HEADER))
                and [header_value]
                or None
            )
            if not cors_domains:
                cors_domains = (
                    (header_value := request.headers.get(HOST_HEADER))
                    and [header_value]
                    or None
                )

    if cors_domains:
        # although the config allows multiple, the header may only contain one.
        # either only one will have been set, or we take the first from the config.
        # this may not be valid in all cases, and so the ASGI CORSMiddleware sould
        # be used instead
        response.headers[CORS_ACCESS_CONTROL_ALLOW_ORIGIN_HEADER] = cors_domains[0]

    response.headers[CORS_ACCESS_CONTROL_ALLOW_CREDENTIALS_HEADER] = str(
        config.web.security.cors.allow_credentials
    )

    response.headers[CORS_ACCESS_CONTROL_ALLOW_METHODS_HEADER] = ",".join(
        config.web.security.cors.allow_methods
    )

    response.headers[CORRELATION_ID_HEADER] = correlation_id

    if config.web.security.csp:
        response.headers[CONTENT_SECURITY_POLICY_HEADER] = config.web.security.csp

    # if config.flask and config.flask.openapi and config.flask.openapi.use_swagger:
    #    # Use a permissive CSP for the Swagger UI
    #    # https://github.com/swagger-api/swagger-ui/issues/7540
    #    FIXME what to use other than `request` w/ Connexion middleware?
    #    if request.path.startswith("/ui/") or (
    #        request.url_rule and request.url_rule.endpoint == "/v1./v1_swagger_ui_index"
    #    ):
    #        response.headers[
    #            CONTENT_SECURITY_POLICY_HEADER
    #        ] = "default-src 'self'; style-src 'self' 'unsafe-inline'; img-src 'self'; script-src 'self' 'unsafe-inline'"


def _log_all_api_responses(response: Response, config: Config, log: Logger):
    correlation_id = _get_correlation_id(log)

    response_headers_safe: Dict[str, str] = dict(response.headers)

    if (
        response_headers_safe.get(RESPONSE_COOKIE_HEADER)
        and config.flask
        and config.flask.session
    ):
        response_headers_safe[RESPONSE_COOKIE_HEADER] = re.sub(
            rf"({config.flask.session.cookie.name}=)[^;]+(;|$)",
            r"\1<redacted>\2",
            response_headers_safe[RESPONSE_COOKIE_HEADER],
        )

    log.info(
        OUTGOING_RESPONSE_MESSAGE,
        response.status_code,
        response.status,
        extra={
            "props": {
                "correlation_id": correlation_id,
                "headers": response_headers_safe,
            }
        },
    )


def register_flask_api_request_handlers(app: Flask):
    return bind_requesthandler(app, Flask.before_request)(_log_all_api_requests)


def register_flask_api_response_handlers(app: Flask):
    # TODO consider moving request/response logging to the WSGI app
    # apparently Flask may not call this if unhandled exceptions occur
    return bind_requesthandler(app, Flask.after_request)(_ordered_api_response_handers)



import uuid
from typing import Literal

import pytest
from connexion import FlaskApp
from flask import Flask, abort
from injector import Module
from Ligare.identity.config import Config as RootSSOConfig
from Ligare.identity.config import SSOConfig
from Ligare.platform.dependency_injection import UserLoaderModule
from Ligare.platform.identity import Role, User
from Ligare.programming.config import AbstractConfig
from Ligare.web.application import OpenAPIAppResult
from Ligare.web.config import Config
from Ligare.web.middleware import bind_errorhandler
from Ligare.web.middleware.consts import CORRELATION_ID_HEADER
from Ligare.web.middleware.flask import (
    _get_correlation_id,  # pyright: ignore[reportPrivateUsage]
)
from Ligare.web.middleware.flask import bind_requesthandler
from Ligare.web.testing.create_app import (
    CreateOpenAPIApp,
    OpenAPIClientInjectorConfigurable,
    OpenAPIMockController,
    RequestConfigurable,
)
from mock import MagicMock
from pytest_mock import MockerFixture
from werkzeug.exceptions import BadRequest, HTTPException, Unauthorized


class TestOpenAPIMiddleware(CreateOpenAPIApp):
    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___register_api_response_handlers__sets_correlation_id_response_header_when_not_set_in_request_header(
        self,
        format: Literal["plaintext", "JSON"],
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_config: Config,
        openapi_mock_controller: OpenAPIMockController,
    ):
        openapi_config.logging.format = format
        openapi_mock_controller.begin()
        flask_client = next(openapi_client_configurable(openapi_config))

        response = flask_client.client.get("/")

        assert response.headers[CORRELATION_ID_HEADER]
        _ = uuid.UUID(response.headers[CORRELATION_ID_HEADER])

    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___register_api_response_handlers__sets_correlation_id_response_header_when_set_in_request_header(
        self,
        format: Literal["plaintext", "JSON"],
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_config: Config,
        openapi_mock_controller: OpenAPIMockController,
    ):
        openapi_config.logging.format = format
        openapi_mock_controller.begin()
        flask_client = next(openapi_client_configurable(openapi_config))
        correlation_id = str(uuid.uuid4())
        response = flask_client.client.get(
            "/", headers={CORRELATION_ID_HEADER: correlation_id}
        )

        assert response.headers[CORRELATION_ID_HEADER] == correlation_id

    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___register_api_response_handler__validates_correlation_id_when_set_in_request_headers(
        self,
        format: Literal["plaintext", "JSON"],
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_config: Config,
        openapi_mock_controller: OpenAPIMockController,
    ):
        openapi_config.logging.format = format
        correlation_id = "abc123"
        openapi_mock_controller.begin()
        flask_client = next(openapi_client_configurable(openapi_config))

        response = flask_client.client.get(
            "/", headers={CORRELATION_ID_HEADER: correlation_id}
        )

        assert response.status_code == 500

    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___get_correlation_id__validates_correlation_id_when_set_in_request_headers(
        self,
        format: Literal["plaintext", "JSON"],
        openapi_request_configurable: RequestConfigurable,
        openapi_config: Config,
        openapi_mock_controller: OpenAPIMockController,
    ):
        openapi_config.logging.format = format
        correlation_id = "abc123"
        openapi_mock_controller.begin()
        with openapi_request_configurable(
            openapi_config, {"headers": {CORRELATION_ID_HEADER: correlation_id}}
        ):
            with pytest.raises(
                ValueError, match="^badly formed hexadecimal UUID string$"
            ):
                _ = _get_correlation_id(MagicMock())

    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___get_correlation_id__uses_existing_correlation_id_when_set_in_request_headers(
        self,
        format: Literal["plaintext", "JSON"],
        openapi_request_configurable: RequestConfigurable,
        openapi_config: Config,
        openapi_mock_controller: OpenAPIMockController,
    ):
        openapi_config.logging.format = format
        correlation_id = str(uuid.uuid4())
        openapi_mock_controller.begin()
        with openapi_request_configurable(
            openapi_config, {"headers": {CORRELATION_ID_HEADER: correlation_id}}
        ):
            returned_correlation_id = _get_correlation_id(MagicMock())
            assert correlation_id == returned_correlation_id

    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___get_correlation_id__sets_correlation_id(
        self,
        format: Literal["plaintext", "JSON"],
        openapi_request_configurable: RequestConfigurable,
        openapi_config: Config,
        openapi_mock_controller: OpenAPIMockController,
    ):
        # app = FlaskApp("foo")
        openapi_config.logging.format = format

        openapi_mock_controller.begin()
        with openapi_request_configurable(openapi_config):
            correlation_id = _get_correlation_id(MagicMock())

            assert correlation_id
            _ = uuid.UUID(correlation_id)

    def test__bind_requesthandler__returns_decorated_flask_request_hook(self):
        app = FlaskApp("foo")

        flask_request_hook_mock = MagicMock()

        wrapped_decorator = bind_requesthandler(app.app, flask_request_hook_mock)
        _ = wrapped_decorator(lambda: None)

        assert flask_request_hook_mock.called

    def test__bind_requesthandler__calls_decorated_function_when_app_is_run(
        self,
        openapi_mock_controller: OpenAPIMockController,
    ):
        app = FlaskApp("foo")

        openapi_mock_controller.begin()

        flask_client = app.test_client()

        wrapped_handler_decorator = bind_requesthandler(app.app, Flask.before_request)
        request_handler_mock = MagicMock()
        _ = wrapped_handler_decorator(request_handler_mock)

        _ = flask_client.get("/")

        assert request_handler_mock.called

    @pytest.mark.parametrize(
        "code_or_exception,openapi_mock_controller",
        [(Exception, lambda: 1), (HTTPException, lambda: 1), (401, lambda: 1)],
        indirect=["openapi_mock_controller"],
    )
    def test__bind_errorhandler__binds_flask_errorhandler(
        self,
        code_or_exception: type[Exception] | int,
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        flask_errorhandler_mock = mocker.patch("flask.Flask.errorhandler")

        def client_init_hook(app: OpenAPIAppResult):
            _ = bind_errorhandler(app.app_injector.app, code_or_exception)

        openapi_mock_controller.begin()
        _ = next(openapi_client_configurable(openapi_config, client_init_hook))

        flask_errorhandler_mock.assert_called_with(code_or_exception)

    @pytest.mark.parametrize(
        "code_or_exception_type,expected_exception_type,openapi_mock_controller",
        [
            (
                Exception,
                ZeroDivisionError,
                lambda: 1 / 0,  # 1/0 to raise an exception (any exception)
            ),
            (HTTPException, BadRequest, lambda: abort(400)),
            (401, Unauthorized, lambda: abort(401)),
        ],
        indirect=["openapi_mock_controller"],
    )
    def test__bind_errorhandler__calls_decorated_function_with_correct_error_when_error_occurs_during_request(
        self,
        code_or_exception_type: type[Exception] | int,
        expected_exception_type: type[Exception],
        openapi_mock_controller: OpenAPIMockController,
    ):
        app = FlaskApp("foo")
        application_errorhandler_mock = MagicMock()
        _ = bind_errorhandler(app, code_or_exception_type)(
            application_errorhandler_mock
        )

        openapi_mock_controller.begin()
        app.add_api("foo.yaml")

        flask_client = app.test_client()

        _ = flask_client.get("/", headers={"Host": "localhost:5000"})

        assert application_errorhandler_mock.called
        assert isinstance(
            application_errorhandler_mock.call_args[0][0], expected_exception_type
        )

    # FIXME just a temporary test as an example set up for testing SAML2
    def test__SAML2Middleware__something(
        self,
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        def app_init_hook(
            application_configs: list[type[AbstractConfig]],
            application_modules: list[Module | type[Module]],
        ):
            application_modules.append(
                UserLoaderModule(
                    loader=User,  # pyright: ignore[reportArgumentType]
                    roles=Role,  # pyright: ignore[reportArgumentType]
                    user_table=MagicMock(),  # pyright: ignore[reportArgumentType]
                    role_table=MagicMock(),  # pyright: ignore[reportArgumentType]
                    bases=[],
                )
            )

        def client_init_hook(app: OpenAPIAppResult):
            app.app_injector.flask_injector.injector.binder.bind(
                RootSSOConfig,
                to=RootSSOConfig(
                    sso=SSOConfig(
                        protocol="SAML2",
                        settings={
                            "relay_state": "http://localhost:5000/sso/redir",
                            "metadata_url": "http://example.org",
                            "metadata": "<xml />",
                        },
                    )
                ),
            )

        openapi_mock_controller.begin()
        app = next(
            openapi_client_configurable(openapi_config, client_init_hook, app_init_hook)
        )

        response = app.client.get("/saml/logout")

        # 401 for now because no real auth is configured.
        # if SSO was broken, 500 would return
        assert response.status_code == 401



"""
This type stub file was generated by pyright.
"""

import abc
import typing as t
from connexion.frameworks.abstract import Framework
from connexion.frameworks.starlette import Starlette as StarletteFramework
from connexion.uri_parsing import AbstractURIParser

class BaseDecorator:
    """Base class for connexion decorators."""
    framework: t.Type[Framework]
    def __init__(self, *, pythonic_params: bool = ..., uri_parser_class: AbstractURIParser = ..., jsonifier=...) -> None:
        ...
    
    @property
    def uri_parser(self):
        ...
    
    def decorate(self, function: t.Callable) -> t.Callable:
        """Decorate a function with decorators based on the operation."""
        ...
    
    @abc.abstractmethod
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    


class WSGIDecorator(BaseDecorator):
    """Decorator for usage with WSGI apps. The parameter decorator works with a Flask request,
    and provides Flask datastructures to the view function. This works for any WSGI app, since
    we get the request via the connexion context provided by WSGI middleware.

    This decorator does not parse responses, but passes them directly to the WSGI App."""
    framework = ...
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    


class FlaskDecorator(WSGIDecorator):
    """Decorator for usage with Connexion or Flask apps. The parameter decorator works with a
    Flask request, and provides Flask datastructures to the view function.

    The response decorator returns Flask responses."""
    ...


class ASGIDecorator(BaseDecorator):
    """Decorator for usage with ASGI apps. The parameter decorator works with a Starlette request,
    and provides Starlette datastructures to the view function. This works for any ASGI app, since
    we get the request via the connexion context provided by ASGI middleware.

    This decorator does not parse responses, but passes them directly to the ASGI App."""
    framework = StarletteFramework
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    


class StarletteDecorator(ASGIDecorator):
    """Decorator for usage with Connexion or Starlette apps. The parameter decorator works with a
    Starlette request, and provides Starlette datastructures to the view function.

    The response decorator returns Starlette responses."""
    ...





"""
This type stub file was generated by pyright.
"""

from connexion.operations.abstract import AbstractOperation

"""
This module defines an OpenAPIOperation class, a Connexion operation specific for OpenAPI 3 specs.
"""
logger = ...
class OpenAPIOperation(AbstractOperation):
    """
    A single API operation on a path.
    """
    def __init__(self, method, path, operation, resolver, path_parameters=..., app_security=..., security_schemes=..., components=..., randomize_endpoint=..., uri_parser_class=...) -> None:
        """
        This class uses the OperationID identify the module and function that will handle the operation

        From Swagger Specification:

        **OperationID**

        A friendly name for the operation. The id MUST be unique among all operations described in the API.
        Tools and libraries MAY use the operation id to uniquely identify an operation.

        :param method: HTTP method
        :type method: str
        :param path:
        :type path: str
        :param operation: swagger operation object
        :type operation: dict
        :param resolver: Callable that maps operationID to a function
        :param path_parameters: Parameters defined in the path level
        :type path_parameters: list
        :param app_security: list of security rules the application uses by default
        :type app_security: list
        :param security_schemes: `Security Definitions Object
            <https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#security-definitions-object>`_
        :type security_schemes: dict
        :param components: `Components Object
            <https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#componentsObject>`_
        :type components: dict
        :param randomize_endpoint: number of random characters to append to operation name
        :type randomize_endpoint: integer
        :param uri_parser_class: class to use for uri parsing
        :type uri_parser_class: AbstractURIParser
        """
        ...
    
    @classmethod
    def from_spec(cls, spec, *args, path, method, resolver, **kwargs): # -> Self:
        ...
    
    @property
    def request_body(self):
        ...
    
    @property
    def parameters(self):
        ...
    
    @property
    def consumes(self): # -> list[Any]:
        ...
    
    @property
    def produces(self): # -> list[Any]:
        ...
    
    def with_definitions(self, schema: dict): # -> dict[Any, Any]:
        ...
    
    def response_schema(self, status_code=..., content_type=...): # -> dict[Any, Any]:
        ...
    
    def example_response(self, status_code=..., content_type=...): # -> tuple[Any, int] | tuple[Any | dict[Any, Any] | list[Any], int] | tuple[None, int]:
        """
        Returns example response from spec
        """
        ...
    
    def get_path_parameter_types(self): # -> dict[Any, Any]:
        ...
    
    def body_name(self, _content_type: str) -> str:
        ...
    
    def body_schema(self, content_type: str = ...) -> dict:
        """
        The body schema definition for this operation.
        """
        ...
    
    def body_definition(self, content_type: str = ...) -> dict:
        """
        The body complete definition for this operation.

        **There can be one "body" parameter at most.**
        """
        ...
    





from logging import Logger
from typing import Any


def on_create(config: dict[str, Any], log: Logger):
    """Called when an application is created."""
    log.debug(f"Database module hook executed")

    config["module"]["database"] = {}

    connection_string = input(
        "\nEnter a database connection string.\nBy default this is `sqlite:///:memory:?check_same_thread=False`.\nRetain this default by pressing enter, or type something else.\n> "
    )

    config["module"]["database"]["connection_string"] = (
        connection_string
        if connection_string
        else "sqlite:///:memory:?check_same_thread=False"
    )
    log.info(
        f"Using database connection string `{config['module']['database']['connection_string']}`"
    )



"""
This type stub file was generated by pyright.
"""

import typing as t
from starlette.types import ASGIApp, Receive, Scope, Send
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.operations import AbstractOperation

"""
Validation Middleware.
"""
logger = ...
class ResponseValidationOperation:
    def __init__(self, next_app: ASGIApp, *, operation: AbstractOperation, validator_map: t.Optional[dict] = ...) -> None:
        ...
    
    def extract_content_type(self, headers: t.List[t.Tuple[bytes, bytes]]) -> t.Tuple[str, str]:
        """Extract the mime type and encoding from the content type headers.

        :param headers: Headers from ASGI scope

        :return: A tuple of mime type, encoding
        """
        ...
    
    def validate_mime_type(self, mime_type: str) -> None:
        """Validate the mime type against the spec if it defines which mime types are produced.

        :param mime_type: mime type from content type header
        """
        ...
    
    @staticmethod
    def validate_required_headers(headers: t.List[tuple], response_definition: dict) -> None:
        ...
    
    async def __call__(self, scope: Scope, receive: Receive, send: Send): # -> None:
        ...
    


class ResponseValidationAPI(RoutedAPI[ResponseValidationOperation]):
    """Validation API."""
    def __init__(self, *args, validator_map=..., validate_responses=..., **kwargs) -> None:
        ...
    
    def make_operation(self, operation: AbstractOperation) -> ResponseValidationOperation:
        ...
    


class ResponseValidationMiddleware(RoutedMiddleware[ResponseValidationAPI]):
    """Middleware for validating requests according to the API contract."""
    api_cls = ResponseValidationAPI





"""
This type stub file was generated by pyright.
"""

import abc
import typing as t
from typing_extensions import Protocol

class Framework(Protocol):
    @staticmethod
    @abc.abstractmethod
    def is_framework_response(response: t.Any) -> bool:
        """Return True if provided response is a framework type"""
        ...
    
    @classmethod
    @abc.abstractmethod
    def connexion_to_framework_response(cls, response):
        """Cast ConnexionResponse to framework response class"""
        ...
    
    @classmethod
    @abc.abstractmethod
    def build_response(cls, data: t.Any, *, content_type: str = ..., headers: dict = ..., status_code: int = ...):
        ...
    
    @staticmethod
    @abc.abstractmethod
    def get_request(*args, **kwargs):
        """Return a framework request from the context."""
        ...
    
    @staticmethod
    @abc.abstractmethod
    def get_body(request):
        """Get body from a framework request"""
        ...
    





"""
This type stub file was generated by pyright.
"""

import types
import typing as t

_ReturnType = t.TypeVar("_ReturnType")
MaybeAwaitable = t.Union[t.Awaitable[_ReturnType], _ReturnType]
Environ = t.Mapping[str, object]
_WriteCallable = t.Callable[[bytes], t.Any]
_ExcInfo = t.Tuple[type, BaseException, types.TracebackType]
_StartResponseCallable = t.Callable[[str, t.Sequence[t.Tuple[str, str]]], _WriteCallable,]
_StartResponseCallableWithExcInfo = t.Callable[[str, t.Sequence[t.Tuple[str, str]], t.Optional[_ExcInfo]], _WriteCallable,]
StartResponse = t.Union[_StartResponseCallable, _StartResponseCallableWithExcInfo]
ResponseStream = t.Iterable[bytes]
WSGIApp = t.Callable[[Environ, StartResponse], ResponseStream]



import logging
from typing import Any, cast

import pytest
from flask import Flask, Response
from Ligare.web.config import Config
from Ligare.web.middleware import register_api_request_handlers
from Ligare.web.middleware.consts import (
    CONTENT_SECURITY_POLICY_HEADER,
    CORS_ACCESS_CONTROL_ALLOW_CREDENTIALS_HEADER,
    CORS_ACCESS_CONTROL_ALLOW_METHODS_HEADER,
    CORS_ACCESS_CONTROL_ALLOW_ORIGIN_HEADER,
    OUTGOING_RESPONSE_MESSAGE,
)
from Ligare.web.middleware.flask import bind_requesthandler
from Ligare.web.testing.create_app import (
    CreateFlaskApp,
    FlaskClientInjector,
    FlaskClientInjectorConfigurable,
)
from pytest import LogCaptureFixture
from pytest_mock import MockerFixture


class TestApiResponseHandlers(CreateFlaskApp):
    def test__register_api_response_handlers__binds_flask_before_request(
        self, flask_client: FlaskClientInjector, mocker: MockerFixture
    ):
        flask_before_request_mock = mocker.patch(
            "flask.sansio.scaffold.Scaffold.before_request"
        )

        _ = register_api_request_handlers(flask_client.client.application)

        assert flask_before_request_mock.called

    def test__wrap_all_api_responses__sets_CSP_header(
        self,
        flask_client_configurable: FlaskClientInjectorConfigurable,
        basic_config: Config,
    ):
        csp_value = "default-src 'self' cdn.example.com;"
        basic_config.web.security.csp = csp_value
        flask_client = next(flask_client_configurable(basic_config))
        response = flask_client.client.get("/")
        header_value = response.headers.get(CONTENT_SECURITY_POLICY_HEADER)
        assert header_value == csp_value

    @pytest.mark.parametrize(
        "header,value,config_attribute_name",
        [
            (CORS_ACCESS_CONTROL_ALLOW_ORIGIN_HEADER, ["example.com"], "origins"),
            (
                CORS_ACCESS_CONTROL_ALLOW_CREDENTIALS_HEADER,
                "False",
                "allow_credentials",
            ),
            (
                CORS_ACCESS_CONTROL_ALLOW_CREDENTIALS_HEADER,
                "True",
                "allow_credentials",
            ),
            (CORS_ACCESS_CONTROL_ALLOW_METHODS_HEADER, ["GET"], "allow_methods"),
            (
                CORS_ACCESS_CONTROL_ALLOW_METHODS_HEADER,
                ["POST", "OPTIONS", "GET"],
                "allow_methods",
            ),
        ],
    )
    def test__wrap_all_api_responses__sets_CORS_headers(
        self,
        header: str,
        value: str,
        config_attribute_name: str,
        flask_client_configurable: FlaskClientInjectorConfigurable,
        basic_config: Config,
    ):
        setattr(basic_config.web.security.cors, config_attribute_name, value)
        flask_client = next(flask_client_configurable(basic_config))
        response = flask_client.client.get("/")
        header_value = response.headers.get(header)
        assert header_value == ",".join(value) if isinstance(value, list) else value

    def test__log_all_api_responses__logs_response_information(
        self,
        flask_client: FlaskClientInjector,
        caplog: LogCaptureFixture,
    ):
        with caplog.at_level(logging.DEBUG):
            response = flask_client.client.get("/")

        msg = OUTGOING_RESPONSE_MESSAGE % (
            response.status_code,
            response.status,
        )

        assert msg in {record.message for record in caplog.records}

    @pytest.mark.parametrize("property_name", ["correlation_id", "headers"])
    def test__log_all_api_responses__logs_extra_response_information(
        self,
        property_name: str,
        flask_client: FlaskClientInjector,
        caplog: LogCaptureFixture,
    ):
        with caplog.at_level(logging.DEBUG):
            response = flask_client.client.get("/")

        msg = OUTGOING_RESPONSE_MESSAGE % (
            response.status_code,
            response.status,
        )

        records = [
            cast(dict[str, Any], cast(Any, record).props)
            for record in caplog.records
            if record.message == msg
        ]
        for record_props in records:
            assert property_name in record_props
            assert record_props[property_name]

    def test__log_all_api_responses__logs_response_headers_without_session_id(
        self,
        flask_client: FlaskClientInjector,
        caplog: LogCaptureFixture,
    ):
        wrapped_handler_decorator = bind_requesthandler(
            flask_client.client.application, Flask.before_request
        )

        def set_session_cookie():
            response = flask_client.injector.injector.get(Response)
            response.set_cookie("session", "foo")
            return response

        _ = wrapped_handler_decorator(set_session_cookie)

        with caplog.at_level(logging.DEBUG):
            response = flask_client.client.get("/")
        msg = OUTGOING_RESPONSE_MESSAGE % (
            response.status_code,
            response.status,
        )

        records = [
            cast(dict[str, Any], cast(Any, record).props)
            for record in caplog.records
            if record.message == msg
        ]
        for record_props in records:
            assert "headers" in record_props
            assert "Set-Cookie" in record_props["headers"]
            assert "session=<redacted>" in record_props["headers"]["Set-Cookie"]



"""
This type stub file was generated by pyright.
"""

import abc

"""
This module defines an AbstractOperation class which implements an abstract Operation interface
and functionality shared between Swagger 2 and OpenAPI 3 specifications.
"""
logger = ...
DEFAULT_MIMETYPE = ...
class AbstractOperation(metaclass=abc.ABCMeta):
    """
    An API routes requests to an Operation by a (path, method) pair.
    The operation uses a resolver to resolve its handler function.
    We use the provided spec to do a bunch of heavy lifting before
    (and after) we call security_schemes handler.
    The registered handler function ends up looking something like::

        @secure_endpoint
        @validate_inputs
        @deserialize_function_inputs
        @serialize_function_outputs
        @validate_outputs
        def user_provided_handler_function(important, stuff):
            if important:
                serious_business(stuff)
    """
    def __init__(self, method, path, operation, resolver, app_security=..., security_schemes=..., randomize_endpoint=..., uri_parser_class=...) -> None:
        """
        :param method: HTTP method
        :type method: str
        :param path:
        :type path: str
        :param operation: swagger operation object
        :type operation: dict
        :param resolver: Callable that maps operationID to a function
        :param app_security: list of security rules the application uses by default
        :type app_security: list
        :param security_schemes: `Security Definitions Object
            <https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#security-definitions-object>`_
        :type security_schemes: dict
        :param randomize_endpoint: number of random characters to append to operation name
        :type randomize_endpoint: integer
        :param uri_parser_class: class to use for uri parsing
        :type uri_parser_class: AbstractURIParser
        """
        ...
    
    @classmethod
    @abc.abstractmethod
    def from_spec(cls, spec, *args, path, method, resolver, **kwargs): # -> None:
        ...
    
    @property
    def method(self): # -> Any:
        """
        The HTTP method for this operation (ex. GET, POST)
        """
        ...
    
    @property
    def request_body(self): # -> None:
        """The request body for this operation"""
        ...
    
    @property
    def is_request_body_defined(self) -> bool:
        """Whether the request body is defined for this operation"""
        ...
    
    @property
    def path(self): # -> Any:
        """
        The path of the operation, relative to the API base path
        """
        ...
    
    @property
    def security(self):
        ...
    
    @property
    def security_schemes(self): # -> None:
        ...
    
    @property
    def responses(self):
        """
        Returns the responses for this operation
        """
        ...
    
    @property
    def operation_id(self):
        """
        The operation id used to identify the operation internally to the app
        """
        ...
    
    @property
    def randomize_endpoint(self): # -> None:
        """
        number of random digits to generate and append to the operation_id.
        """
        ...
    
    @property
    def router_controller(self):
        """
        The router controller to use (python module where handler functions live)
        """
        ...
    
    @property
    @abc.abstractmethod
    def parameters(self): # -> None:
        """
        Returns the parameters for this operation
        """
        ...
    
    @property
    @abc.abstractmethod
    def produces(self): # -> None:
        """
        Content-Types that the operation produces
        """
        ...
    
    @property
    @abc.abstractmethod
    def consumes(self): # -> None:
        """
        Content-Types that the operation consumes
        """
        ...
    
    @abc.abstractmethod
    def body_name(self, content_type: str) -> str:
        """
        Name of the body in the spec.
        """
        ...
    
    @abc.abstractmethod
    def body_schema(self, content_type: str = ...) -> dict:
        """
        The body schema definition for this operation.
        """
        ...
    
    @abc.abstractmethod
    def body_definition(self, content_type: str = ...) -> dict:
        """
        The body definition for this operation.
        :rtype: dict
        """
        ...
    
    def response_definition(self, status_code=..., content_type=...):
        """
        response definition for this endpoint
        """
        ...
    
    @abc.abstractmethod
    def response_schema(self, status_code=..., content_type=...): # -> None:
        """
        response schema for this endpoint
        """
        ...
    
    @abc.abstractmethod
    def example_response(self, status_code=..., content_type=...): # -> None:
        """
        Returns an example from the spec
        """
        ...
    
    @abc.abstractmethod
    def get_path_parameter_types(self): # -> None:
        """
        Returns the types for parameters in the path
        """
        ...
    
    @abc.abstractmethod
    def with_definitions(self, schema): # -> None:
        """
        Returns the given schema, but with the definitions from the spec
        attached. This allows any remaining references to be resolved by a
        validator (for example).
        """
        ...
    
    def get_mimetype(self): # -> Literal['application/json']:
        """
        If the endpoint has no 'produces' then the default is
        'application/json'.

        :rtype str
        """
        ...
    
    @property
    def uri_parser_class(self): # -> None:
        """
        The uri parser class for this operation.
        """
        ...
    
    @property
    def function(self):
        """
        Resolved function.

        :rtype: types.FunctionType
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

NoContent = ...
class MediaTypeDict(dict):
    """
    A dictionary where keys can be either media types or media type ranges. When fetching a
    value from the dictionary, the provided key is checked against the ranges. The most specific
    key is chosen as prescribed by the OpenAPI spec, with `type/*` being preferred above
    `*/subtype`.
    """
    def __getitem__(self, item):
        ...
    
    def get(self, item, default=...): # -> None:
        ...
    
    def __contains__(self, item): # -> bool:
        ...
    





"""
This type stub file was generated by pyright.
"""

import enum
import pathlib
import typing as t
from dataclasses import dataclass

from connexion.jsonifier import Jsonifier
from connexion.lifecycle import ConnexionRequest, ConnexionResponse
from connexion.middleware.lifespan import Lifespan
from connexion.options import SwaggerUIOptions
from connexion.resolver import Resolver
from connexion.types import MaybeAwaitable
from connexion.uri_parsing import AbstractURIParser
from starlette.types import ASGIApp, Receive, Scope, Send

logger = ...

@dataclass
class _Options:
    """
    Connexion provides a lot of parameters for the user to configure the app / middleware of
    application.

    This class provides a central place to parse these parameters a mechanism to update them.
    Application level arguments can be provided when instantiating the application / middleware,
    after which they can be overwritten on an API level.

    The defaults should only be set in this class, and set to None in the signature of user facing
    methods. This is necessary for this class to be able to differentiate between missing and
    falsy arguments.
    """

    arguments: t.Optional[dict] = ...
    auth_all_paths: t.Optional[bool] = ...
    jsonifier: t.Optional[Jsonifier] = ...
    pythonic_params: t.Optional[bool] = ...
    resolver: t.Optional[t.Union[Resolver, t.Callable]] = ...
    resolver_error: t.Optional[int] = ...
    resolver_error_handler: t.Optional[t.Callable] = ...
    strict_validation: t.Optional[bool] = ...
    swagger_ui_options: t.Optional[SwaggerUIOptions] = ...
    uri_parser_class: t.Optional[AbstractURIParser] = ...
    validate_responses: t.Optional[bool] = ...
    validator_map: t.Optional[dict] = ...
    security_map: t.Optional[dict] = ...
    def __post_init__(self):  # -> None:
        ...
    def replace(self, **changes) -> _Options:
        """Update mechanism to overwrite the options. None values are discarded.

        :param changes: Arguments accepted by the __init__ method of this class.

        :return: An new _Options object with updated arguments.
        """
        ...

class MiddlewarePosition(enum.Enum):
    """Positions to insert a middleware"""

    BEFORE_EXCEPTION = ...
    BEFORE_SWAGGER = ...
    BEFORE_ROUTING = ...
    BEFORE_SECURITY = ...
    BEFORE_VALIDATION = ...
    BEFORE_CONTEXT = ...

class API:
    def __init__(self, specification, *, base_path, **kwargs) -> None: ...

class ConnexionMiddleware:
    """The main Connexion middleware, which wraps a list of specialized middlewares around the
    provided application."""

    default_middlewares = ...
    middlewares: list[type] | None = ...
    def __init__(
        self,
        app: ASGIApp,
        *,
        import_name: t.Optional[str] = ...,
        lifespan: t.Optional[Lifespan] = ...,
        middlewares: t.Optional[t.List[ASGIApp]] = ...,
        specification_dir: t.Union[pathlib.Path, str] = ...,
        arguments: t.Optional[dict] = ...,
        auth_all_paths: t.Optional[bool] = ...,
        jsonifier: t.Optional[Jsonifier] = ...,
        pythonic_params: t.Optional[bool] = ...,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = ...,
        resolver_error: t.Optional[int] = ...,
        strict_validation: t.Optional[bool] = ...,
        swagger_ui_options: t.Optional[SwaggerUIOptions] = ...,
        uri_parser_class: t.Optional[AbstractURIParser] = ...,
        validate_responses: t.Optional[bool] = ...,
        validator_map: t.Optional[dict] = ...,
        security_map: t.Optional[dict] = ...,
    ) -> None:
        """
        :param import_name: The name of the package or module that this object belongs to. If you
            are using a single module, __name__ is always the correct value. If you however are
            using a package, it’s usually recommended to hardcode the name of your package there.
        :param middlewares: The list of middlewares to wrap around the application. Defaults to
            :obj:`middleware.main.ConnexionmMiddleware.default_middlewares`
        :param specification_dir: The directory holding the specification(s). The provided path
            should either be absolute or relative to the root path of the application. Defaults to
            the root path.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: Instance of :class:`options.ConnexionOptions` with
            configuration options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`.
        :param security_map: A dictionary of security handlers to use. Defaults to
            :obj:`security.SECURITY_HANDLERS`.
        """
        ...
    def add_middleware(
        self,
        middleware_class: t.Type[ASGIApp] | t.Any,
        *,
        position: MiddlewarePosition = ...,
        **options: t.Any,
    ) -> None:
        """Add a middleware to the stack on the specified position.

        :param middleware_class: Middleware class to add
        :param position: Position to add the middleware, one of the MiddlewarePosition Enum
        :param options: Options to pass to the middleware_class on initialization
        """
        ...
    def add_api(
        self,
        specification: t.Union[pathlib.Path, str, dict],
        *,
        base_path: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        arguments: t.Optional[dict] = ...,
        auth_all_paths: t.Optional[bool] = ...,
        jsonifier: t.Optional[Jsonifier] = ...,
        pythonic_params: t.Optional[bool] = ...,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = ...,
        resolver_error: t.Optional[int] = ...,
        strict_validation: t.Optional[bool] = ...,
        swagger_ui_options: t.Optional[SwaggerUIOptions] = ...,
        uri_parser_class: t.Optional[AbstractURIParser] = ...,
        validate_responses: t.Optional[bool] = ...,
        validator_map: t.Optional[dict] = ...,
        security_map: t.Optional[dict] = ...,
        **kwargs,
    ) -> None:
        """
        Register een API represented by a single OpenAPI specification on this middleware.
        Multiple APIs can be registered on a single middleware.

        :param specification: OpenAPI specification. Can be provided either as dict, or as path
            to file.
        :param base_path: Base path to host the API. This overrides the basePath / servers in the
            specification.
        :param name: Name to register the API with. If no name is passed, the base_path is used
            as name instead.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: A dict with configuration options for the swagger ui. See
            :class:`options.ConnexionOptions`.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`
        :param security_map: A dictionary of security handlers to use. Defaults to
            :obj:`security.SECURITY_HANDLERS`
        :param kwargs: Additional keyword arguments to pass to the `add_api` method of the managed
            middlewares. This can be used to pass arguments to middlewares added beyond the default
            ones.

        :return: The Api registered on the wrapped application.
        """
        ...
    def add_error_handler(
        self,
        code_or_exception: t.Union[int, t.Type[Exception]],
        function: t.Callable[
            [ConnexionRequest, Exception], MaybeAwaitable[ConnexionResponse]
        ],
    ) -> None:
        """
        Register a callable to handle application errors.

        :param code_or_exception: An exception class or the status code of HTTP exceptions to
            handle.
        :param function: Callable that will handle exception, may be async.
        """
        ...
    def run(self, import_string: str = ..., **kwargs):  # -> None:
        """Run the application using uvicorn.

        :param import_string: application as import string (eg. "main:app"). This is needed to run
                              using reload.
        :param kwargs: kwargs to pass to `uvicorn.run`.
        """
        ...
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None: ...
    middleware_stack: t.Optional[t.Iterable[ASGIApp]] = None



import logging
from typing import Any, cast

import pytest
from Ligare.web.middleware import register_api_request_handlers
from Ligare.web.middleware.consts import INCOMING_REQUEST_MESSAGE
from Ligare.web.testing.create_app import CreateFlaskApp, FlaskClientInjector
from pytest import LogCaptureFixture
from pytest_mock import MockerFixture


class TestApiRequestHandlers(CreateFlaskApp):
    def test__register_api_request_handlers__binds_flask_before_request(
        self, flask_client: FlaskClientInjector, mocker: MockerFixture
    ):
        flask_before_request_mock = mocker.patch(
            "flask.sansio.scaffold.Scaffold.before_request"
        )

        _ = register_api_request_handlers(flask_client.client.application)

        assert flask_before_request_mock.called

    def test__log_all_api_requests__logs_request_information(
        self,
        flask_client: FlaskClientInjector,
        caplog: LogCaptureFixture,
    ):
        with caplog.at_level(logging.DEBUG):
            response = flask_client.client.get("/")

        request = response.request
        msg = INCOMING_REQUEST_MESSAGE % (
            request.method,
            request.url,
            request.host,
            request.remote_addr,
            request.remote_user,
        )

        assert msg in {record.message for record in caplog.records}

    @pytest.mark.parametrize("property_name", ["correlation_id", "headers"])
    def test__log_all_api_requests__logs_extra_request_information(
        self,
        property_name: str,
        flask_client: FlaskClientInjector,
        caplog: LogCaptureFixture,
    ):
        with caplog.at_level(logging.DEBUG):
            response = flask_client.client.get("/")

        request = response.request
        msg = INCOMING_REQUEST_MESSAGE % (
            request.method,
            request.url,
            request.host,
            request.remote_addr,
            request.remote_user,
        )

        records = [
            cast(dict[str, Any], cast(Any, record).props)
            for record in caplog.records
            if record.message == msg
        ]
        for record_props in records:
            assert property_name in record_props
            assert record_props[property_name]

    def test__log_all_api_requests__logs_request_headers_without_session_id(
        self,
        flask_client: FlaskClientInjector,
        caplog: LogCaptureFixture,
    ):
        with caplog.at_level(logging.DEBUG):
            response = flask_client.client.get("/")

        request = response.request
        msg = INCOMING_REQUEST_MESSAGE % (
            request.method,
            request.url,
            request.host,
            request.remote_addr,
            request.remote_user,
        )

        records = [
            cast(dict[str, Any], cast(Any, record).props)
            for record in caplog.records
            if record.message == msg
        ]
        for record_props in records:
            assert "headers" in record_props
            assert "Cookie" in record_props["headers"]
            assert "session=<redacted>" in record_props["headers"]["Cookie"]



import pytest
from flask import Flask
from Ligare.programming.collections.dict import AnyDict
from Ligare.programming.config import AbstractConfig, load_config
from Ligare.programming.config.exceptions import ConfigBuilderStateError
from Ligare.web.application import ApplicationBuilder, ApplicationConfigBuilder
from Ligare.web.config import Config
from Ligare.web.exception import BuilderBuildError, InvalidBuilderStateError
from pydantic import BaseModel
from pytest_mock import MockerFixture
from typing_extensions import override


def test__Config__load_config__reads_toml_file(mocker: MockerFixture):
    fake_config_dict = {}
    toml_mock = mocker.patch("toml.load", return_value=fake_config_dict)
    _ = load_config(Config, "foo.toml")
    assert toml_mock.called


def test__Config__load_config__initializes_logging_config(mocker: MockerFixture):
    fake_config_dict = {"logging": {"log_level": "DEBUG"}}
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads", return_value=fake_config_dict)
    config = load_config(Config, "foo.toml")
    assert config.logging.log_level == "DEBUG"


def test__Config__load_config__initializes_flask_config(mocker: MockerFixture):
    fake_config_dict = {"flask": {"app_name": "test_app"}}
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads", return_value=fake_config_dict)
    config = load_config(Config, "foo.toml")
    assert config.flask is not None
    assert config.flask.app_name == "test_app"


def test__Config__prepare_env_for_flask__requires_flask_secret_key_when_sessions_are_used(
    mocker: MockerFixture,
):
    fake_config_dict: AnyDict = {
        "flask": {"app_name": "test_app", "session": {"cookie": {}}}
    }
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads", return_value=fake_config_dict)

    with pytest.raises(
        Exception, match=r"^`flask.session.cookie.secret_key` must be set in config.$"
    ):
        _ = load_config(Config, "foo.toml")


@pytest.mark.parametrize("mode", ["ssm", "filename"])
def test__ApplicationConfigBuilder__build__succeeds_with_either_ssm_or_filename(
    mode: str, mocker: MockerFixture
):
    fake_config_dict = {"logging": {"log_level": "DEBUG"}}
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads", return_value=fake_config_dict)
    _ = mocker.patch("Ligare.AWS.ssm.SSMParameters.load_config")

    application_config_builder = ApplicationConfigBuilder[
        Config
    ]().with_root_config_type(Config)

    if mode == "ssm":
        _ = application_config_builder.enable_ssm(True)
    else:
        _ = application_config_builder.with_config_filename("foo.toml")

    _ = application_config_builder.build()


def test__ApplicationConfigBuilder__build__raises_InvalidBuilderStateError_without_ssm_or_filename(
    mocker: MockerFixture,
):
    fake_config_dict = {"logging": {"log_level": "DEBUG"}}
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads", return_value=fake_config_dict)
    _ = mocker.patch("Ligare.AWS.ssm.SSMParameters.load_config")

    application_config_builder = ApplicationConfigBuilder[Config]()

    with pytest.raises(InvalidBuilderStateError):
        _ = application_config_builder.build()


def test__ApplicationConfigBuilder__build__raises_BuilderBuildError_when_ssm_fails_and_filename_not_configured(
    mocker: MockerFixture,
):
    fake_config_dict = {"logging": {"log_level": "DEBUG"}}
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads", return_value=fake_config_dict)
    _ = mocker.patch(
        "Ligare.AWS.ssm.SSMParameters.load_config",
        side_effect=Exception("Test mode failure."),
    )

    application_config_builder = ApplicationConfigBuilder[Config]().enable_ssm(True)

    with pytest.raises(BuilderBuildError):
        _ = application_config_builder.build()


def test__ApplicationConfigBuilder__build__uses_filename_when_ssm_fails(
    mocker: MockerFixture,
):
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads")
    _ = mocker.patch(
        "Ligare.AWS.ssm.SSMParameters.load_config",
        side_effect=Exception("Test mode failure."),
    )
    toml_mock = mocker.patch("toml.load")

    application_config_builder = (
        ApplicationConfigBuilder[Config]()
        .with_root_config_type(Config)
        .with_config_filename("foo.toml")
    )

    _ = application_config_builder.build()

    assert toml_mock.called


def test__ApplicationConfigBuilder__build__requires_root_config(
    mocker: MockerFixture,
):
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads")

    application_config_builder = ApplicationConfigBuilder[
        Config
    ]().with_config_filename("foo.toml")

    with pytest.raises(
        BuilderBuildError,
        match="A root config must be specified",
    ) as e:
        _ = application_config_builder.build()

    assert isinstance(e.value.__cause__, ConfigBuilderStateError)


def test__ApplicationConfigBuilder__build__applies_additional_configs(
    mocker: MockerFixture,
):
    fake_config_dict = {"logging": {"log_level": "DEBUG"}, "test": {"foo": "bar"}}
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads", return_value=fake_config_dict)

    class TestConfig(BaseModel, AbstractConfig):
        @override
        def post_load(self) -> None:
            return super().post_load()

        foo: str

    application_config_builder = (
        ApplicationConfigBuilder[Config]()
        .with_root_config_type(Config)
        .with_config_type(TestConfig)
        .with_config_filename("foo.toml")
    )
    config = application_config_builder.build()

    assert config is not None
    assert hasattr(config, "test")
    assert hasattr(getattr(config, "test"), "foo")
    assert getattr(getattr(config, "test"), "foo") == "bar"


def test__ApplicationConfigBuilder__build__applies_config_overrides(
    mocker: MockerFixture,
):
    fake_config_dict = {"logging": {"log_level": "DEBUG"}}
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads", return_value=fake_config_dict)

    application_config_builder = (
        ApplicationConfigBuilder[Config]()
        .with_root_config_type(Config)
        .with_config_value_overrides({"logging": {"log_level": "INFO"}})
        .with_config_filename("foo.toml")
    )
    config = application_config_builder.build()

    assert config is not None
    assert hasattr(config, "logging")
    assert hasattr(getattr(config, "logging"), "log_level")
    assert getattr(getattr(config, "logging"), "log_level") == "INFO"


# FIXME move to application tests
def test__ApplicationBuilder__build__something(mocker: MockerFixture):
    fake_config_dict = {"logging": {"log_level": "DEBUG"}, "flask": {"app_name": "app"}}
    _ = mocker.patch("io.open")
    _ = mocker.patch("toml.decoder.loads", return_value=fake_config_dict)

    application_builder = ApplicationBuilder[Flask]().use_configuration(
        lambda config_builder: config_builder.with_root_config_type(
            Config
        ).with_config_filename("foo.toml")
    )

    _ = (
        application_builder.with_flask_app_name("overridden_app")
        .with_flask_env("overridden_dev")
        .build()
    )



"""
This type stub file was generated by pyright.
"""

import dataclasses
import typing as t

"""
This module defines a Connexion specific options class to pass to the Connexion App or API.
"""
NO_UI_MSG = ...
logger = ...
@dataclasses.dataclass
class SwaggerUIOptions:
    """Options to configure the Swagger UI.

    :param serve_spec: Whether to serve the Swagger / OpenAPI Specification
    :param spec_path: Where to serve the Swagger / OpenAPI Specification

    :param swagger_ui: Whether to serve the Swagger UI
    :param swagger_ui_path: Where to serve the Swagger UI
    :param swagger_ui_config: Options to configure the Swagger UI. See
          https://swagger.io/docs/open-source-tools/swagger-ui/usage/configuration
          for an overview of the available options.
    :param swagger_ui_template_dir: Directory with static files to use to serve Swagger UI
    :param swagger_ui_template_arguments: Arguments passed to the Swagger UI template. Useful
        when providing your own template dir with additional template arguments.
    """
    serve_spec: bool = ...
    spec_path: t.Optional[str] = ...
    swagger_ui: bool = ...
    swagger_ui_config: dict = ...
    swagger_ui_path: str = ...
    swagger_ui_template_dir: t.Optional[str] = ...
    swagger_ui_template_arguments: dict = ...


class SwaggerUIConfig:
    """Class holding swagger UI specific options."""
    def __init__(self, options: t.Optional[SwaggerUIOptions] = ..., oas_version: t.Tuple[int, ...] = ...) -> None:
        ...
    
    @property
    def openapi_spec_available(self) -> bool:
        """Whether to make the OpenAPI Specification available."""
        ...
    
    @property
    def openapi_spec_path(self) -> str:
        """Path to host the Swagger UI."""
        ...
    
    @property
    def swagger_ui_available(self) -> bool:
        """Whether to make the Swagger UI available."""
        ...
    
    @property
    def swagger_ui_path(self) -> str:
        """Path to mount the Swagger UI and make it accessible via a browser."""
        ...
    
    @property
    def swagger_ui_template_dir(self) -> str:
        """Directory with static files to use to serve Swagger UI."""
        ...
    
    @property
    def swagger_ui_config(self) -> dict:
        """Options to configure the Swagger UI."""
        ...
    
    @property
    def swagger_ui_template_arguments(self) -> dict:
        """Arguments passed to the Swagger UI template."""
        ...
    





"""
This type stub file was generated by pyright.
"""

from contextvars import ContextVar
from starlette.types import Receive, Scope
from connexion.operations import AbstractOperation

UNBOUND_MESSAGE = ...
_context: ContextVar[dict] = ...
context = ...
_operation: ContextVar[AbstractOperation] = ...
operation = ...
_receive: ContextVar[Receive] = ...
receive = ...
_scope: ContextVar[Scope] = ...
scope = ...
request = ...



from os import environ

import pytest
from connexion import FlaskApp
from flask import Flask
from Ligare.programming.config import AbstractConfig
from Ligare.programming.str import get_random_str
from Ligare.web.application import App  # pyright:ignore[reportDeprecated]
from Ligare.web.application import ApplicationBuilder, configure_openapi
from Ligare.web.config import Config, FlaskConfig, FlaskOpenApiConfig
from Ligare.web.testing.create_app import CreateOpenAPIApp
from mock import MagicMock
from pydantic import BaseModel
from pytest_mock import MockerFixture
from typing_extensions import override


class TestCreateOpenAPIApp(CreateOpenAPIApp):
    def test__CreateOpenAPIApp__configure_openapi__requires_flask_config(self):
        with pytest.raises(
            Exception,
            match=r"^OpenAPI configuration is empty\. Review the `openapi` section of your application's `config\.toml`\.$",
        ):
            _ = configure_openapi(Config())

    def test__CreateOpenAPIApp__configure_openapi__creates_flask_app_using_config(
        self, mocker: MockerFixture
    ):
        _ = mocker.patch("Ligare.web.application.json_logging")
        connexion_mock = mocker.patch("Ligare.web.application.FlaskApp")

        app_name = f"{TestCreateOpenAPIApp.test__CreateOpenAPIApp__configure_openapi__creates_flask_app_using_config.__name__}-app_name"
        spec_path = ".."

        _ = configure_openapi(
            Config(
                flask=FlaskConfig(
                    app_name=app_name,
                    openapi=FlaskOpenApiConfig(spec_path=spec_path),
                )
            )
        )

        connexion_mock.assert_called_with(app_name, specification_dir=spec_path)

    def test__CreateOpenAPIApp__create_app__loads_config_from_toml(
        self, basic_config: Config, mocker: MockerFixture
    ):
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        load_config_mock = mocker.patch(
            "Ligare.web.application.load_config", return_value=basic_config
        )

        toml_filename = f"{TestCreateOpenAPIApp.test__CreateOpenAPIApp__create_app__loads_config_from_toml.__name__}-config.toml"
        application_builder = ApplicationBuilder[Flask]().use_configuration(
            lambda config_builder: config_builder.with_config_filename(toml_filename)
        )
        _ = application_builder.build()
        assert load_config_mock.called
        assert load_config_mock.call_args and load_config_mock.call_args[0]
        assert load_config_mock.call_args[0][1] == toml_filename

    def test__CreateOpenAPIApp__create_app__uses_custom_config_types(
        self, mocker: MockerFixture
    ):
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        toml_filename = f"{TestCreateOpenAPIApp.test__CreateOpenAPIApp__create_app__uses_custom_config_types.__name__}-config.toml"
        toml_load_result = {
            "flask": {
                "app_name": f"{TestCreateOpenAPIApp.test__CreateOpenAPIApp__create_app__uses_custom_config_types.__name__}-app_name"
            },
            "custom": {"foo": get_random_str(k=26)},
        }

        _ = mocker.patch("toml.load", return_value=toml_load_result)

        class CustomConfig(BaseModel, AbstractConfig):
            @override
            def post_load(self) -> None:
                return super().post_load()

            foo: str = get_random_str(k=26)

        app = App[Flask].create(  # pyright:ignore[reportDeprecated]
            config_filename=toml_filename, application_configs=[CustomConfig]
        )

        assert (
            app.app_injector.flask_injector.injector.get(CustomConfig).foo
            == toml_load_result["custom"]["foo"]
        )

    @pytest.mark.parametrize(
        "envvar_name,config_var_name,var_value",
        [
            ("FLASK_APP", "app_name", "foobar"),
            ("FLASK_ENV", "env", "barfoo"),
        ],
    )
    def test__CreateOpenAPIApp__create_app__updates_flask_config_from_envvars(
        self,
        envvar_name: str,
        config_var_name: str,
        var_value: str,
        basic_config: Config,
        mocker: MockerFixture,
    ):
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        object.__setattr__(basic_config.flask, config_var_name, var_value)

        environ.update({envvar_name: var_value})
        _ = mocker.patch(
            "Ligare.web.application.load_config", return_value=basic_config
        )
        _ = App[Flask].create()  # pyright:ignore[reportDeprecated]

        assert object.__getattribute__(basic_config.flask, config_var_name) == var_value

    @pytest.mark.parametrize(
        "envvar_name,config_var_name,var_value,should_fail",
        [
            ("FLASK_APP", None, "foobar", False),
            ("FLASK_ENV", None, "barfoo", False),
            (None, "app_name", "foobar", False),
            (None, "app_name", "", True),
            (None, "env", "barfoo", False),
        ],
    )
    def test__CreateOpenAPIApp__create_app__requires_application_name(
        self,
        envvar_name: str | None,
        config_var_name: str | None,
        var_value: str,
        should_fail: bool,
        mocker: MockerFixture,
    ):
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        environ.update({"FLASK_APP": "", "FLASK_ENV": ""})

        if envvar_name is not None:
            environ.update({envvar_name: var_value})

        toml_load_result = {}
        if config_var_name is not None:
            toml_load_result["flask"] = {config_var_name: var_value}

        _ = mocker.patch("toml.load", return_value=toml_load_result)

        if should_fail:
            with pytest.raises(Exception):
                _ = App[Flask].create()  # pyright:ignore[reportDeprecated]
        else:
            _ = App[Flask].create()  # pyright:ignore[reportDeprecated]

    def test__CreateOpenAPIApp__create_app__configures_appropriate_app_type_based_on_config(
        self, mocker: MockerFixture
    ):
        toml_filename = f"{TestCreateOpenAPIApp.test__CreateOpenAPIApp__create_app__configures_appropriate_app_type_based_on_config.__name__}-config.toml"
        app_name = f"{TestCreateOpenAPIApp.test__CreateOpenAPIApp__create_app__configures_appropriate_app_type_based_on_config.__name__}-app_name"
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        _ = mocker.patch("Ligare.web.application.register_error_handlers")
        _ = mocker.patch("Ligare.web.application.register_api_request_handlers")
        _ = mocker.patch("Ligare.web.application.register_api_response_handlers")
        _ = mocker.patch("Ligare.web.application.configure_dependencies")

        configure_method_mock = mocker.patch("Ligare.web.application.configure_openapi")
        config = Config(
            flask=FlaskConfig(app_name=app_name, openapi=FlaskOpenApiConfig())
        )
        _ = mocker.patch("Ligare.web.application.load_config", return_value=config)
        _ = App[FlaskApp].create(config_filename=toml_filename)  # pyright:ignore[reportDeprecated]

        configure_method_mock.assert_called_once_with(config)



import pathlib
from os import environ
from pathlib import Path
from typing import cast

import pytest
from flask import Blueprint, Flask
from Ligare.programming.config import AbstractConfig
from Ligare.programming.str import get_random_str
from Ligare.web.application import App  # pyright: ignore[reportDeprecated]
from Ligare.web.application import configure_blueprint_routes
from Ligare.web.config import Config, FlaskConfig
from Ligare.web.testing.create_app import (
    CreateFlaskApp,
    FlaskClientInjectorConfigurable,
)
from mock import MagicMock
from pydantic import BaseModel
from pytest_mock import MockerFixture
from typing_extensions import override


class TestCreateFlaskApp(CreateFlaskApp):
    def test__CreateFlaskApp__configure_blueprint_routes__requires_flask_config(self):
        with pytest.raises(
            Exception,
            match=r"^Flask configuration is empty\. Review the `flask` section of your application's `config\.toml`\.$",
        ):
            _ = configure_blueprint_routes(Config())

    def test__CreateFlaskApp__configure_blueprint_routes__creates_flask_app_using_config(
        self, mocker: MockerFixture
    ):
        _ = mocker.patch("Ligare.web.application.json_logging")
        flask_mock = mocker.patch("Ligare.web.application.Flask")

        app_name = f"{TestCreateFlaskApp.test__CreateFlaskApp__configure_blueprint_routes__creates_flask_app_using_config.__name__}-app_name"

        _ = configure_blueprint_routes(Config(flask=FlaskConfig(app_name=app_name)))

        flask_mock.assert_called_with(app_name)

    @pytest.mark.parametrize("filename", ["foo", "foo.py", "__main__.py"])
    def test__CreateFlaskApp__configure_blueprint_routes__when_discovering_blueprints_ignores_directories_in_path(
        self, filename: str, mocker: MockerFixture
    ):
        mocker.stop(
            self._automatic_mocks["Ligare.web.application._import_blueprint_modules"]
        )

        spec_lookup_mock = mocker.patch("importlib.util.spec_from_file_location")
        glob_item_mock = MagicMock(
            is_file=MagicMock(
                # fakes a directory
                return_value=False
            )
        )
        type(glob_item_mock).name = filename

        app_name = f"{TestCreateFlaskApp.test__CreateFlaskApp__configure_blueprint_routes__when_discovering_blueprints_ignores_directories_in_path.__name__}-app_name"

        _path = Path
        _ = mocker.patch(
            "pathlib.Path",
            return_value=MagicMock(
                glob=MagicMock(return_value=[glob_item_mock]),
            ),
        )

        try:
            flask_app = configure_blueprint_routes(
                Config(
                    flask=FlaskConfig(app_name=app_name),
                ),
                ".",
            )
        finally:
            # mocker doesn't appear to restore pathlib.Path correctly
            # either through automatic cleanup, `with` statements,
            # or `mocker.stop`.
            pathlib.Path = _path

        assert not spec_lookup_mock.called
        assert flask_app.blueprints == {}

    @pytest.mark.parametrize(
        "is_file,filename",
        [
            (True, "foo"),
            (True, "foo.py"),
            (True, "__main__.py"),
            (True, "__init__.py"),
            (False, "__init__.py"),
        ],
    )
    def test__CreateFlaskApp__configure_blueprint_routes__when_discovering_blueprints_registers_python_files_and_modules(
        self, is_file: bool, filename: str, mocker: MockerFixture
    ):
        mocker.stop(
            self._automatic_mocks["Ligare.web.application._import_blueprint_modules"]
        )

        glob_item_mock = MagicMock(
            is_file=MagicMock(
                # fakes a directory
                return_value=is_file
            )
        )
        type(glob_item_mock).name = filename

        spec_lookup_mock = mocker.patch("importlib.util.spec_from_file_location")
        _ = mocker.patch("importlib.util.module_from_spec")

        app_name = f"{TestCreateFlaskApp.test__CreateFlaskApp__configure_blueprint_routes__when_discovering_blueprints_registers_python_files_and_modules.__name__}-app_name"

        _ = mocker.patch(
            "pathlib.Path",
            return_value=MagicMock(
                glob=MagicMock(return_value=[glob_item_mock]),
            ),
        )

        _path = Path
        try:
            flask_app = configure_blueprint_routes(
                Config(
                    flask=FlaskConfig(app_name=app_name),
                ),
                ".",
            )
        finally:
            # mocker doesn't appear to restore pathlib.Path correctly
            # either through automatic cleanup, `with` statements,
            # or `mocker.stop`.
            pathlib.Path = _path

        spec_lookup_mock.assert_called()
        assert flask_app.blueprints == {}

    @pytest.mark.parametrize(
        "is_file,filename",
        [
            (True, "foo"),
            (True, "foo.py"),
            (True, "__main__.py"),
            (True, "__init__.py"),
            (False, "__init__.py"),
        ],
    )
    def test__CreateFlaskApp__configure_blueprint_routes__when_discovering_blueprints_registers_blueprint_modules(
        self, is_file: bool, filename: str, mocker: MockerFixture
    ):
        mocker.stop(
            self._automatic_mocks["Ligare.web.application._import_blueprint_modules"]
        )
        _ = mocker.patch("importlib.util.spec_from_file_location")
        _ = mocker.patch("flask.app.Flask.register_blueprint")

        glob_item_mock = MagicMock(
            is_file=MagicMock(
                # fakes a directory
                return_value=is_file
            )
        )
        type(glob_item_mock).name = filename

        blueprint = Blueprint("foo_blueprint", "")
        _ = mocker.patch(
            "importlib.util.module_from_spec",
            return_value=MagicMock(foo_blueprint=blueprint),
        )

        app_name = f"{TestCreateFlaskApp.test__CreateFlaskApp__configure_blueprint_routes__when_discovering_blueprints_registers_blueprint_modules.__name__}-app_name"

        _path = Path
        _ = mocker.patch(
            "pathlib.Path",
            return_value=MagicMock(
                glob=MagicMock(return_value=[glob_item_mock, glob_item_mock]),
            ),
        )

        try:
            flask_app = configure_blueprint_routes(
                Config(
                    flask=FlaskConfig(app_name=app_name),
                ),
                ".",
            )
        finally:
            # mocker doesn't appear to restore pathlib.Path correctly
            # either through automatic cleanup, `with` statements,
            # or `mocker.stop`.
            pathlib.Path = _path

        assert cast(MagicMock, flask_app.register_blueprint).call_count == 2
        cast(MagicMock, flask_app.register_blueprint).assert_called_with(blueprint)

    def test__CreateFlaskApp__configure_blueprint_routes__when_discovering_blueprints_stops_when_module_load_fails(
        self, mocker: MockerFixture
    ):
        mocker.stop(
            self._automatic_mocks["Ligare.web.application._import_blueprint_modules"]
        )

        glob_item_mock = MagicMock(
            is_file=MagicMock(
                # fakes a directory
                return_value=True
            )
        )
        type(glob_item_mock).name = "__init__.py"
        _ = mocker.patch("importlib.util.spec_from_file_location", return_value=None)
        _ = mocker.patch("importlib.util.module_from_spec")

        _path = Path
        _ = mocker.patch(
            "pathlib.Path",
            return_value=MagicMock(
                glob=MagicMock(return_value=[glob_item_mock]),
            ),
        )
        app_name = f"{TestCreateFlaskApp.test__CreateFlaskApp__configure_blueprint_routes__when_discovering_blueprints_stops_when_module_load_fails.__name__}-app_name"

        with pytest.raises(
            Exception,
            match=rf"^Module cannot be created from path {glob_item_mock}$",
        ):
            try:
                _ = configure_blueprint_routes(
                    Config(
                        flask=FlaskConfig(app_name=app_name),
                    ),
                    ".",
                )
            finally:
                # mocker doesn't appear to restore pathlib.Path correctly
                # either through automatic cleanup, `with` statements,
                # or `mocker.stop`.
                pathlib.Path = _path

    def test__CreateFlaskApp__create_app__requires_flask_config(
        self, flask_client_configurable: FlaskClientInjectorConfigurable
    ):
        with pytest.raises(
            Exception,
            match=r"^You must set \[flask\] in the application configuration\.$",
        ):
            _ = next(flask_client_configurable(Config()))

    def test__CreateFlaskApp__create_app__loads_config_from_toml(
        self, basic_config: Config, mocker: MockerFixture
    ):
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        load_config_mock = mocker.patch(
            "Ligare.web.application.load_config", return_value=basic_config
        )

        toml_filename = f"{TestCreateFlaskApp.test__CreateFlaskApp__create_app__loads_config_from_toml.__name__}-config.toml"
        _ = App[Flask].create(config_filename=toml_filename)  # pyright: ignore[reportDeprecated]
        assert load_config_mock.called
        assert load_config_mock.call_args and load_config_mock.call_args[0]
        assert load_config_mock.call_args[0][1] == toml_filename

    def test__CreateFlaskApp__create_app__uses_custom_config_types(
        self, mocker: MockerFixture
    ):
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        toml_filename = f"{TestCreateFlaskApp.test__CreateFlaskApp__create_app__uses_custom_config_types.__name__}-config.toml"
        toml_load_result = {
            "flask": {
                "app_name": f"{TestCreateFlaskApp.test__CreateFlaskApp__create_app__uses_custom_config_types.__name__}-app_name"
            },
            "custom": {"foo": get_random_str(k=26)},
        }

        _ = mocker.patch("toml.load", return_value=toml_load_result)

        class CustomConfig(BaseModel, AbstractConfig):
            @override
            def post_load(self) -> None:
                return super().post_load()

            foo: str = get_random_str(k=26)

        app = App[Flask].create(  # pyright: ignore[reportDeprecated]
            config_filename=toml_filename, application_configs=[CustomConfig]
        )

        assert (
            app.app_injector.flask_injector.injector.get(CustomConfig).foo
            == toml_load_result["custom"]["foo"]
        )

    @pytest.mark.parametrize(
        "envvar_name,config_var_name,var_value",
        [
            ("FLASK_APP", "app_name", "foobar"),
            ("FLASK_ENV", "env", "barfoo"),
        ],
    )
    def test__CreateFlaskApp__create_app__updates_flask_config_from_envvars(
        self,
        envvar_name: str,
        config_var_name: str,
        var_value: str,
        basic_config: Config,
        mocker: MockerFixture,
    ):
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        object.__setattr__(basic_config.flask, config_var_name, var_value)

        environ.update({envvar_name: var_value})
        _ = mocker.patch(
            "Ligare.web.application.load_config", return_value=basic_config
        )
        _ = App[Flask].create()  # pyright: ignore[reportDeprecated]

        assert object.__getattribute__(basic_config.flask, config_var_name) == var_value

    @pytest.mark.parametrize(
        "envvar_name,config_var_name,var_value,should_fail",
        [
            ("FLASK_APP", None, "foobar", False),
            ("FLASK_ENV", None, "barfoo", False),
            (None, "app_name", "foobar", False),
            (None, "app_name", "", True),
            (None, "env", "barfoo", False),
        ],
    )
    def test__CreateFlaskApp__create_app__requires_application_name(
        self,
        envvar_name: str | None,
        config_var_name: str | None,
        var_value: str,
        should_fail: bool,
        mocker: MockerFixture,
    ):
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        environ.update({"FLASK_APP": "", "FLASK_ENV": ""})

        if envvar_name is not None:
            environ.update({envvar_name: var_value})

        toml_load_result = {}
        if config_var_name is not None:
            toml_load_result["flask"] = {config_var_name: var_value}

        _ = mocker.patch("toml.load", return_value=toml_load_result)

        if should_fail:
            with pytest.raises(Exception):
                _ = App[Flask].create()  # pyright: ignore[reportDeprecated]
        else:
            _ = App[Flask].create()  # pyright: ignore[reportDeprecated]

    def test__CreateFlaskApp__create_app__configures_appropriate_app_type_based_on_config(
        self, mocker: MockerFixture
    ):
        toml_filename = f"{TestCreateFlaskApp.test__CreateFlaskApp__create_app__configures_appropriate_app_type_based_on_config.__name__}-config.toml"
        app_name = f"{TestCreateFlaskApp.test__CreateFlaskApp__create_app__configures_appropriate_app_type_based_on_config.__name__}-app_name"
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        _ = mocker.patch("Ligare.web.application.register_error_handlers")
        _ = mocker.patch("Ligare.web.application.register_api_request_handlers")
        _ = mocker.patch("Ligare.web.application.register_api_response_handlers")
        _ = mocker.patch("Ligare.web.application.configure_dependencies")

        configure_method_mock = mocker.patch(
            "Ligare.web.application.configure_blueprint_routes"
        )
        config = Config(flask=FlaskConfig(app_name=app_name))
        _ = mocker.patch("Ligare.web.application.load_config", return_value=config)
        _ = App[Flask].create(config_filename=toml_filename)  # pyright: ignore[reportDeprecated]

        configure_method_mock.assert_called_once_with(config)



"""
This type stub file was generated by pyright.
"""

from .__about__ import __version__
from .config import AUTH_HEADER_NAME, COOKIE_DURATION, COOKIE_HTTPONLY, COOKIE_NAME, COOKIE_SECURE, ID_ATTRIBUTE, LOGIN_MESSAGE, LOGIN_MESSAGE_CATEGORY, REFRESH_MESSAGE, REFRESH_MESSAGE_CATEGORY
from .login_manager import LoginManager
from .mixins import AnonymousUserMixin, UserMixin
from .signals import session_protected, user_accessed, user_loaded_from_cookie, user_loaded_from_request, user_logged_in, user_logged_out, user_login_confirmed, user_needs_refresh, user_unauthorized
from .test_client import FlaskLoginClient
from .utils import confirm_login, current_user, decode_cookie, encode_cookie, fresh_login_required, login_fresh, login_remembered, login_required, login_url, login_user, logout_user, make_next_param, set_login_view

__all__ = ["__version__", "AUTH_HEADER_NAME", "COOKIE_DURATION", "COOKIE_HTTPONLY", "COOKIE_NAME", "COOKIE_SECURE", "ID_ATTRIBUTE", "LOGIN_MESSAGE", "LOGIN_MESSAGE_CATEGORY", "REFRESH_MESSAGE", "REFRESH_MESSAGE_CATEGORY", "LoginManager", "AnonymousUserMixin", "UserMixin", "session_protected", "user_accessed", "user_loaded_from_cookie", "user_loaded_from_request", "user_logged_in", "user_logged_out", "user_login_confirmed", "user_needs_refresh", "user_unauthorized", "FlaskLoginClient", "confirm_login", "current_user", "decode_cookie", "encode_cookie", "fresh_login_required", "login_fresh", "login_remembered", "login_required", "login_url", "login_user", "logout_user", "make_next_param", "set_login_view"]
def __getattr__(name): # -> NamedSignal:
    ...




"""
This type stub file was generated by pyright.
"""

__title__ = ...
__description__ = ...
__url__ = ...
__version_info__ = ...
__version__ = ...
__author__ = ...
__author_email__ = ...
__maintainer__ = ...
__license__ = ...
__copyright__ = ...



"""
This type stub file was generated by pyright.
"""

import flask_restful
import flask_restx
from injector import Scope, __version__ as injector_version
from flask.views import View

def test_injections(): # -> None:
    class View2(View):
        ...
    
    

def test_resets(): # -> None:
    class OurScope(Scope):
        ...
    
    

def test_memory_leak(): # -> None:
    ...

def test_doesnt_raise_deprecation_warning(): # -> None:
    ...

def test_jinja_env_globals_support_injection(): # -> None:
    ...

def test_error_handlers_support_injection(): # -> None:
    class CustomException(Exception):
        ...
    
    

def test_view_functions_arent_modified_globally(): # -> None:
    class MyView(View):
        ...
    
    

def test_view_args_and_class_args_are_passed_to_class_based_views(): # -> None:
    class MyView(View):
        ...
    
    

def test_flask_restful_integration_works(): # -> None:
    class HelloWorld(flask_restful.Resource):
        ...
    
    

def test_flask_restx_integration_works(): # -> None:
    class HelloWorld(flask_restx.Resource):
        ...
    
    

def test_request_scope_not_started_before_any_request_made_regression(): # -> None:
    ...

def test_noninstrospectable_hooks_dont_crash_everything(): # -> None:
    ...

def test_instance_methods(): # -> None:
    class HelloWorldService:
        ...
    
    
    class HelloWorld:
        ...
    
    

if injector_version >= '0.12':
    def test_forward_references_work(): # -> None:
        class X:
            ...
        
        
    
def test_request_scope_covers_teardown_request_handlers(): # -> None:
    ...

def test_request_scope_covers_blueprint_teardown_request_handlers(): # -> None:
    ...




"""
This type stub file was generated by pyright.
"""

import typing as t
import yaml
from jsonschema.exceptions import ValidationError

"""
Module containing all code related to json schema validation.
"""
class ExtendedSafeLoader(yaml.SafeLoader):
    """Extends the yaml SafeLoader to coerce all keys to string so the result is valid json."""
    def __init__(self, stream) -> None:
        ...
    
    def extended_construct_mapping(self, node, deep=...): # -> dict[str, Any]:
        ...
    


class FileHandler:
    """Handler to resolve file refs."""
    def __call__(self, uri): # -> Any:
        ...
    


class URLHandler:
    """Handler to resolve url refs."""
    def __call__(self, uri): # -> Any:
        ...
    


handlers = ...
def resolve_refs(spec, store=..., base_uri=...): # -> Mapping[Any, Any] | list[Any] | tuple[Any, ...]:
    """
    Resolve JSON references like {"$ref": <some URI>} in a spec.
    Optionally takes a store, which is a mapping from reference URLs to a
    dereferenced objects. Prepopulating the store can avoid network calls.
    """
    ...

def format_error_with_path(exception: ValidationError) -> str:
    """Format a `ValidationError` with path to error."""
    ...

def allow_nullable(validation_fn: t.Callable) -> t.Callable:
    """Extend an existing validation function, so it allows nullable values to be null."""
    ...

def validate_writeOnly(validator, wo, instance, schema): # -> Generator[ValidationError, Any, None]:
    ...

NullableTypeValidator = ...
NullableEnumValidator = ...
Draft4RequestValidator = ...
Draft4ResponseValidator = ...



"""
This type stub file was generated by pyright.
"""

import sys

def is_env_var_toggle(var_name): # -> bool:
    ...

def get_library_logger(logger_name): # -> Logger:
    """

    :param logger_name: name
    :return: logger
    """
    ...

def update_formatter_for_loggers(loggers_iter, formatter): # -> None:
    """
    :param formatter:
    :param loggers_iter:
    """
    ...

def parse_int(input_int, default): # -> int:
    ...

def validate_subclass(subclass, superclass): # -> Literal[True]:
    """

    :param subclass
    :param superclass
    :return: bool
    """
    ...

_epoch = ...
def epoch_nano_second(datetime_):
    ...

def iso_time_format(datetime_): # -> str:
    ...

if hasattr(sys, '_getframe'):
    currentframe = ...
else:
    def currentframe(_no_of_go_up_level):
        """Return the frame object for the caller's stack frame."""
        ...
    
class RequestUtil:
    """
        util for extract request's information
    """
    def __new__(cls, *args, **kw): # -> Self@RequestUtil:
        ...
    
    def get_correlation_id(self, request=..., within_formatter=...): # -> str:
        """
        Gets the correlation id from the header of the request. \
        It tries to search from json_logging.CORRELATION_ID_HEADERS list, one by one.\n
        If found no value, new id will be generated by default.\n
        :param request: request object
        :return: correlation id string
        """
        ...
    
    def get_request_from_call_stack(self, within_formatter=...): # -> Any | None:
        """

        :return: get request object from call stack
        """
        ...
    


def is_not_match_any_pattern(path, patterns): # -> bool:
    ...




"""
This type stub file was generated by pyright.
"""

import typing as t
from starlette.requests import Request as StarletteRequest
from werkzeug import Request as WerkzeugRequest

"""
This module defines interfaces for requests and responses used in Connexion for authentication,
validation, serialization, etc.
"""
class _RequestInterface:
    @property
    def context(self) -> t.Dict[str, t.Any]:
        """The connexion context of the current request cycle."""
        ...
    
    @property
    def content_type(self) -> str:
        """The content type included in the request headers."""
        ...
    
    @property
    def mimetype(self) -> str:
        """The content type included in the request headers stripped from any optional character
        set encoding"""
        ...
    
    @property
    def path_params(self) -> t.Dict[str, t.Any]:
        """Path parameters exposed as a dictionary"""
        ...
    
    @property
    def query_params(self) -> t.Dict[str, t.Any]:
        """Query parameters exposed as a dictionary"""
        ...
    
    def form(self) -> t.Union[t.Dict[str, t.Any], t.Awaitable[t.Dict[str, t.Any]]]:
        """Form data, including files."""
        ...
    
    def files(self) -> t.Dict[str, t.Any]:
        """Files included in the request."""
        ...
    
    def json(self) -> dict:
        """Json data included in the request."""
        ...
    
    def get_body(self) -> t.Any:
        """Get body based on the content type. This returns json data for json content types,
        form data for form content types, and bytes for all others. If the bytes data is empty,
        :code:`None` is returned instead."""
        ...
    


class WSGIRequest(_RequestInterface):
    def __init__(self, werkzeug_request: WerkzeugRequest, uri_parser=..., view_args=...) -> None:
        ...
    
    @property
    def context(self): # -> Any:
        ...
    
    @property
    def content_type(self) -> str:
        ...
    
    @property
    def mimetype(self) -> str:
        ...
    
    @property
    def path_params(self):
        ...
    
    @property
    def query_params(self):
        ...
    
    def form(self):
        ...
    
    def files(self): # -> dict[str, list[FileStorage]]:
        ...
    
    def json(self): # -> Any:
        ...
    
    def get_body(self): # -> Any | None:
        ...
    
    def __getattr__(self, item): # -> Any:
        ...
    


class ConnexionRequest(_RequestInterface):
    """
    Implementation of the Connexion :code:`_RequestInterface` representing an ASGI request.

    .. attribute:: _starlette_request
        :noindex:

        This class wraps a Starlette `Request <https://www.starlette.io/requests/#request>`_,
        and provides access to its attributes by proxy.

    """
    def __init__(self, *args, uri_parser=..., **kwargs) -> None:
        ...
    
    @classmethod
    def from_starlette_request(cls, request: StarletteRequest, uri_parser=...) -> ConnexionRequest:
        ...
    
    @property
    def context(self): # -> Any:
        ...
    
    @property
    def content_type(self): # -> Any:
        ...
    
    @property
    def mimetype(self): # -> str:
        ...
    
    @property
    def path_params(self) -> t.Dict[str, t.Any]:
        ...
    
    @property
    def query_params(self):
        ...
    
    async def form(self): # -> None:
        ...
    
    async def files(self): # -> defaultdict[Any, list[Any]] | None:
        ...
    
    async def json(self): # -> Any | None:
        ...
    
    async def get_body(self): # -> Any | None:
        ...
    
    def __getattr__(self, item): # -> Any | None:
        ...
    


class ConnexionResponse:
    """Connexion interface for a response."""
    def __init__(self, status_code=..., mimetype=..., content_type=..., body=..., headers=..., is_streamed=...) -> None:
        ...
    





"""
This type stub file was generated by pyright.
"""

import abc
import typing as t
from connexion.frameworks.abstract import Framework

logger = ...
class BaseResponseDecorator:
    def __init__(self, *, framework: t.Type[Framework], jsonifier) -> None:
        ...
    
    @abc.abstractmethod
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    
    def build_framework_response(self, handler_response):
        ...
    


class SyncResponseDecorator(BaseResponseDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    


class AsyncResponseDecorator(BaseResponseDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    


class NoResponseDecorator(BaseResponseDecorator):
    """Dummy decorator to skip response serialization."""
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    





"""
This type stub file was generated by pyright.
"""

import typing as t
import yaml
from connexion.middleware.main import API

"""
This module provides general utility functions used within Connexion.
"""
if t.TYPE_CHECKING:
    ...
def boolean(s): # -> bool:
    """
    Convert JSON/Swagger boolean value to Python, raise ValueError otherwise

    >>> boolean('true')
    True

    >>> boolean('false')
    False
    """
    ...

TYPE_MAP: t.Dict[str, t.Any] = ...
def make_type(value: t.Any, type_: str, format_: t.Optional[str]) -> t.Any:
    """Cast a value to the type defined in the specification."""
    ...

def deep_merge(a, b):
    """merges b into a
    in case of conflict the value from b is used
    """
    ...

def deep_getattr(obj, attr): # -> Any:
    """
    Recurses through an attribute chain to get the ultimate value.
    """
    ...

def deep_get(obj, keys):
    """
    Recurses through a nested object get a leaf value.

    There are cases where the use of inheritance or polymorphism-- the use of allOf or
    oneOf keywords-- will cause the obj to be a list. In this case the keys will
    contain one or more strings containing integers.

    :type obj: list or dict
    :type keys: list of strings
    """
    ...

def get_function_from_name(function_name): # -> Any:
    """
    Tries to get function by fully qualified name (e.g. "mymodule.myobj.myfunc")

    :type function_name: str
    """
    ...

def is_json_mimetype(mimetype): # -> Literal[False]:
    """
    :type mimetype: str
    :rtype: bool
    """
    ...

def all_json(mimetypes): # -> bool:
    """
    Returns True if all mimetypes are serialized with json

    :type mimetypes: list
    :rtype: bool

    >>> all_json(['application/json'])
    True
    >>> all_json(['application/x.custom+json'])
    True
    >>> all_json([])
    True
    >>> all_json(['application/xml'])
    False
    >>> all_json(['text/json'])
    False
    >>> all_json(['application/json', 'other/type'])
    False
    >>> all_json(['application/json', 'application/x.custom+json'])
    True
    """
    ...

def is_nullable(param_def):
    ...

def is_null(value): # -> bool:
    ...

def has_coroutine(function, api=...): # -> bool:
    """
    Checks if function is a coroutine.
    If ``function`` is a decorator (has a ``__wrapped__`` attribute)
    this function will also look at the wrapped function.
    """
    ...

def yamldumper(openapi): # -> str:
    """
    Returns a nicely-formatted yaml spec.
    :param openapi: a spec dictionary.
    :return: a nicely-formatted, serialized yaml spec.
    """
    class NoAnchorDumper(yaml.dumper.SafeDumper):
        """A yaml Dumper that does not replace duplicate entries
        with yaml anchors.
        """
        ...
    
    

def not_installed_error(exc, *, msg=...): # -> Callable[..., NoReturn]:
    """Raises the ImportError when the module/object is actually called with a custom message."""
    ...

def extract_content_type(headers: t.Union[t.List[t.Tuple[bytes, bytes]], t.Dict[str, str]]) -> t.Optional[str]:
    """Extract the mime type and encoding from the content type headers.

    :param headers: Headers from ASGI scope

    :return: The content type if available in headers, otherwise None
    """
    ...

def split_content_type(content_type: t.Optional[str]) -> t.Tuple[t.Optional[str], t.Optional[str]]:
    """Split the content type in mime_type and encoding. Other parameters are ignored."""
    ...

def coerce_type(param, value, parameter_type, parameter_name=...): # -> list[Any] | dict[Any, Any] | None:
    ...

def get_root_path(import_name: str) -> str:
    """Copied from Flask:
    https://github.com/pallets/flask/blob/836866dc19218832cf02f8b04911060ac92bfc0b/src/flask/helpers.py#L595

    Find the root path of a package, or the path that contains a
    module. If it cannot be found, returns the current working
    directory.
    """
    ...

def inspect_function_arguments(function: t.Callable) -> t.Tuple[t.List[str], bool]:
    """
    Returns the list of variables names of a function and if it
    accepts keyword arguments.
    """
    ...

T = t.TypeVar("T")
@t.overload
def sort_routes(routes: t.List[str], *, key: None = ...) -> t.List[str]:
    ...

@t.overload
def sort_routes(routes: t.List[T], *, key: t.Callable[[T], str]) -> t.List[T]:
    ...

def sort_routes(routes, *, key=...): # -> list[Any]:
    """Sorts a list of routes from most specific to least specific.

    See Starlette routing documentation and implementation as this function
    is aimed to sort according to that logic.
    - https://www.starlette.io/routing/#route-priority

    The only difference is that a `path` component is appended to each route
    such that `/` is less specific than `/basepath` while they are technically
    not comparable.
    This is because it is also done by the `Mount` class internally:
    https://github.com/encode/starlette/blob/1c1043ca0ab7126419948b27f9d0a78270fd74e6/starlette/routing.py#L388

    For example, from most to least specific:
    - /users/me
    - /users/{username}/projects/{project}
    - /users/{username}

    :param routes: List of routes to sort
    :param key: Function to extract the path from a route if it is not a string

    :return: List of routes sorted from most specific to least specific
    """
    class SortableRoute:
        ...
    
    

def sort_apis_by_basepath(apis: t.List[API]) -> t.List[API]:
    """Sorts a list of APIs by basepath.

    :param apis: List of APIs to sort

    :return: List of APIs sorted by basepath
    """
    ...




from dataclasses import dataclass
from functools import wraps
from logging import Logger
from typing import Any, Callable, Generic, ParamSpec, Sequence, TypedDict, TypeVar, cast

from connexion import FlaskApp, request
from flask import Blueprint, Flask, abort
from injector import Binder, Injector, Module, inject, provider, singleton
from Ligare.platform.feature_flag.caching_feature_flag_router import (
    CachingFeatureFlagRouter,
)
from Ligare.platform.feature_flag.caching_feature_flag_router import (
    FeatureFlag as CachingFeatureFlag,
)
from Ligare.platform.feature_flag.db_feature_flag_router import DBFeatureFlagRouter
from Ligare.platform.feature_flag.db_feature_flag_router import (
    FeatureFlag as DBFeatureFlag,
)
from Ligare.platform.feature_flag.db_feature_flag_router import (
    FeatureFlagTable,
    FeatureFlagTableBase,
)
from Ligare.platform.feature_flag.feature_flag_router import (
    FeatureFlag,
    FeatureFlagRouter,
    TFeatureFlag,
)
from Ligare.programming.config import AbstractConfig
from Ligare.programming.patterns.dependency_injection import ConfigurableModule
from Ligare.web.middleware.sso import login_required
from pydantic import BaseModel
from starlette.types import ASGIApp, Receive, Scope, Send
from typing_extensions import override


class FeatureFlagConfig(BaseModel):
    api_base_url: str = "/platform"
    access_role_name: str | bool | None = None


class Config(BaseModel, AbstractConfig):
    @override
    def post_load(self) -> None:
        return super().post_load()

    feature_flag: FeatureFlagConfig


class FeatureFlagPatchRequest(TypedDict):
    name: str
    enabled: bool


@dataclass
class FeatureFlagPatch:
    name: str
    enabled: bool


class FeatureFlagRouterModule(ConfigurableModule, Generic[TFeatureFlag]):
    def __init__(self, t_feature_flag: type[FeatureFlagRouter[TFeatureFlag]]) -> None:
        self._t_feature_flag = t_feature_flag
        super().__init__()

    @override
    @staticmethod
    def get_config_type() -> type[AbstractConfig]:
        return Config

    @singleton
    @provider
    def _provide_feature_flag_router(
        self, injector: Injector
    ) -> FeatureFlagRouter[FeatureFlag]:
        return injector.get(self._t_feature_flag)


class DBFeatureFlagRouterModule(FeatureFlagRouterModule[DBFeatureFlag]):
    def __init__(self) -> None:
        super().__init__(DBFeatureFlagRouter)

    @singleton
    @provider
    def _provide_db_feature_flag_router(
        self, injector: Injector
    ) -> FeatureFlagRouter[DBFeatureFlag]:
        return injector.get(self._t_feature_flag)

    @singleton
    @provider
    def _provide_db_feature_flag_router_table_base(self) -> type[FeatureFlagTableBase]:
        # FeatureFlagTable is a FeatureFlagTableBase provided through
        # SQLAlchemy's declarative meta API
        return cast(type[FeatureFlagTableBase], FeatureFlagTable)


class CachingFeatureFlagRouterModule(FeatureFlagRouterModule[CachingFeatureFlag]):
    def __init__(self) -> None:
        super().__init__(CachingFeatureFlagRouter)

    @singleton
    @provider
    def _provide_caching_feature_flag_router(
        self, injector: Injector
    ) -> FeatureFlagRouter[CachingFeatureFlag]:
        return injector.get(self._t_feature_flag)


P = ParamSpec("P")
R = TypeVar("R")


@inject
def _get_feature_flag_blueprint(app: Flask, config: FeatureFlagConfig, log: Logger):
    feature_flag_blueprint = Blueprint(
        "feature_flag", __name__, url_prefix=f"{config.api_base_url}"
    )

    access_role = config.access_role_name

    def _login_required(require_flask_login: bool):
        """
        Decorate an API endpoint with the correct flask_login authentication
        method given the requirements of the API endpoint.

        require_flask_login is ignored if flask_login has been configured.

        If flask_login has _not_ been configured:
            * If require_flask_login is True, a warning is logged and the request is aborted with a 405 error
            * If require_flask_login is False, the endpoint function is executed

        :param bool require_flask_login: Determine whether flask_login must be configured for this endpoint to function
        :return _type_: _description_
        """

        def __login_required(
            fn: Callable[P, R],
        ) -> Callable[P, R]:
            authorization_implementation: Callable[..., Any]

            if access_role is False:
                authorization_implementation = fn
            # None means no roles were specified, but a session is still required
            elif access_role is None or access_role is True:
                authorization_implementation = login_required(fn)
            else:
                authorization_implementation = login_required([access_role])(fn)

            @wraps(fn)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                if not hasattr(app, "login_manager"):
                    if require_flask_login:
                        log.warning(
                            "The Feature Flag module expects flask_login to be configured in order to control access to feature flag modifications. flask_login has not been configured, so the Feature Flag modification API is disabled."
                        )

                        return abort(405)
                    else:
                        return fn(*args, **kwargs)
                return authorization_implementation(*args, **kwargs)

            return wrapper

        return __login_required

    @feature_flag_blueprint.route("/feature_flag", methods=("GET",))
    @_login_required(False)
    @inject
    def feature_flag(feature_flag_router: FeatureFlagRouter[FeatureFlag]):  # pyright: ignore[reportUnusedFunction]
        request_query_names: list[str] | None = request.query_params.getlist("name")

        feature_flags: Sequence[FeatureFlag]
        missing_flags: set[str] | None = None
        if request_query_names is None or not request_query_names:
            feature_flags = feature_flag_router.get_feature_flags()
        elif isinstance(request_query_names, list):  # pyright: ignore[reportUnnecessaryIsInstance]
            feature_flags = feature_flag_router.get_feature_flags(request_query_names)
            missing_flags = set(request_query_names).difference(
                set([feature_flag.name for feature_flag in feature_flags])
            )
        else:
            raise ValueError("Unexpected type from Flask query parameters.")

        response: dict[str, Any] = {}
        problems: list[Any] = []

        if missing_flags is not None:
            for missing_flag in missing_flags:
                problems.append({
                    "title": "feature flag not found",
                    "detail": "Queried feature flag does not exist.",
                    "instance": missing_flag,
                    "status": 404,
                    "type": None,
                })
            response["problems"] = problems

        elif not feature_flags:
            problems.append({
                "title": "No feature flags found",
                "detail": "Queried feature flags do not exist.",
                "instance": "",
                "status": 404,
                "type": None,
            })
            response["problems"] = problems

        if feature_flags:
            response["data"] = feature_flags
            return response
        else:
            return response, 404

    @feature_flag_blueprint.route("/feature_flag", methods=("PATCH",))
    @_login_required(True)
    @inject
    async def feature_flag_patch(feature_flag_router: FeatureFlagRouter[FeatureFlag]):  # pyright: ignore[reportUnusedFunction]
        feature_flags_request: list[FeatureFlagPatchRequest] = await request.json()

        feature_flags = [
            FeatureFlagPatch(name=flag["name"], enabled=flag["enabled"])
            for flag in feature_flags_request
        ]

        changes: list[Any] = []
        problems: list[Any] = []
        for flag in feature_flags:
            try:
                change = feature_flag_router.set_feature_is_enabled(
                    flag.name, flag.enabled
                )
                changes.append(change)
            except LookupError:
                problems.append({
                    "title": "feature flag not found",
                    "detail": "Feature flag to PATCH does not exist. It must be created first.",
                    "instance": flag.name,
                    "status": 404,
                    "type": None,
                })

        response: dict[str, Any] = {}

        if problems:
            response["problems"] = problems

        if changes:
            response["data"] = changes
            return response
        else:
            return response, 404

    return feature_flag_blueprint


class FeatureFlagMiddlewareModule(Module):
    """
    Enable the use of Feature Flags and a Feature Flag management API.
    """

    @override
    def configure(self, binder: Binder) -> None:
        super().configure(binder)

    def register_middleware(self, app: FlaskApp):
        app.add_middleware(FeatureFlagMiddlewareModule.FeatureFlagMiddleware)

    class FeatureFlagMiddleware:
        """
        ASGI middleware for Feature Flags.

        This middleware create a Flask blueprint the enables a Feature Flag management API.
        """

        _app: ASGIApp

        def __init__(self, app: ASGIApp):
            super().__init__()
            self._app = app

        @inject
        async def __call__(
            self,
            scope: Scope,
            receive: Receive,
            send: Send,
            app: Flask,
            injector: Injector,
            log: Logger,
        ) -> None:
            async def wrapped_send(message: Any) -> None:
                # Only run during startup of the application
                if (
                    scope["type"] != "lifespan"
                    or message["type"] != "lifespan.startup.complete"
                    or not scope["app"]
                ):
                    return await send(message)

                log.debug("Registering FeatureFlag blueprint.")
                app.register_blueprint(
                    injector.call_with_injection(_get_feature_flag_blueprint)
                )
                log.debug("FeatureFlag blueprint registered.")

                return await send(message)

            await self._app(scope, receive, wrapped_send)



"""
This type stub file was generated by pyright.
"""

from .abstract import SpecMiddleware
from .main import ConnexionMiddleware, MiddlewarePosition

__all__ = ("SpecMiddleware", "ConnexionMiddleware", "MiddlewarePosition")



"""
This type stub file was generated by pyright.
"""

from starlette.types import ASGIApp, Receive, Scope, Send
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.operations import AbstractOperation

"""The ContextMiddleware creates a global context based the scope. It should be last in the
middleware stack, so it exposes the scope passed to the application"""
class ContextOperation:
    def __init__(self, next_app: ASGIApp, *, operation: AbstractOperation) -> None:
        ...
    
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        ...
    


class ContextAPI(RoutedAPI[ContextOperation]):
    def __init__(self, *args, **kwargs) -> None:
        ...
    
    def make_operation(self, operation: AbstractOperation) -> ContextOperation:
        ...
    


class ContextMiddleware(RoutedMiddleware[ContextAPI]):
    """Middleware to expose operation specific context to application."""
    api_cls = ContextAPI





"""
This type stub file was generated by pyright.
"""

COOKIE_NAME = ...
COOKIE_DURATION = ...
COOKIE_SECURE = ...
COOKIE_HTTPONLY = ...
COOKIE_SAMESITE = ...
LOGIN_MESSAGE = ...
LOGIN_MESSAGE_CATEGORY = ...
REFRESH_MESSAGE = ...
REFRESH_MESSAGE_CATEGORY = ...
ID_ATTRIBUTE = ...
AUTH_HEADER_NAME = ...
SESSION_KEYS = ...
EXEMPT_METHODS = ...
USE_SESSION_FOR_NEXT = ...



"""
This type stub file was generated by pyright.
"""

import functools
from inspect import ismethod
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    TypeVar,
    Union,
    cast,
    get_type_hints,
)

import flask
import flask_restx
from flask import Config, Request
from flask_restful import Api as FlaskRestfulApi
from flask_restful.utils import unpack as flask_response_unpack
from flask_restx.utils import unpack as flask_response_unpack
from injector import (
    Binder,
    Injector,
    Module,
    Provider,
    Scope,
    ScopeDecorator,
    inject,
    singleton,
)
from werkzeug.local import Local, LocalManager, LocalProxy
from werkzeug.wrappers import Response

__author__ = ...
__version__ = ...
__all__ = ['request', 'RequestScope', 'Config', 'Request', 'FlaskInjector']

request: ScopeDecorator

T = TypeVar('T', LocalProxy[Any], Callable[..., Any])
def instance_method_wrapper(im: T) -> T:
    ...

def wrap_fun(fun: T, injector: Injector) -> T:
    ...

def wrap_function(fun: Callable[..., Any], injector: Injector) -> Callable[..., Any]:
    ...

def wrap_class_based_view(fun: Callable[..., Any], injector: Injector) -> Callable[..., Any]:
    ...

def wrap_flask_restful_resource(fun: Callable[..., Any], flask_restful_api: Any, injector: Injector) -> Callable[..., Any]:
    """
    This is needed because of how flask_restful views are registered originally.

    :type flask_restful_api: :class:`flask_restful.Api`
    """
    ...

TProvider = TypeVar("TProvider")
class CachedProviderWrapper(Provider[TProvider]):
    def __init__(self, old_provider: Provider[TProvider]) -> None:
        ...
    
    def get(self, injector: Injector) -> Any:
        ...
    


class RequestScope(Scope):
    """A scope whose object lifetime is tied to a request.

    @request
    class Session:
        pass
    """
    if False:
        ...
    def cleanup(self) -> None:
        ...
    
    def prepare(self) -> None:
        ...
    
    def configure(self) -> None:
        ...
    
    def get(self, key: Any, provider: Provider[TProvider]) -> Any:
        ...
    
    _locals: Local
    _local_manager: LocalManager


request = ...
_ModuleT = Union[Callable[[Binder], Any], Module]
class FlaskInjector:
    injector: Injector
    app: flask.Flask
    def __init__(self, app: flask.Flask, modules: Iterable[_ModuleT] = ..., injector: Injector = ..., request_scope_class: type = ...) -> None:
        """Initializes Injector for the application.

        .. note::

            Needs to be called *after* all views, signal handlers, template globals
            and context processors are registered.

        :param app: Application to configure
        :param modules: Configuration for newly created :class:`injector.Injector`
        :param injector: Injector to initialize app with, if not provided
            a new instance will be created.
        :type app: :class:`flask.Flask`
        :type modules: Iterable of configuration modules
        :rtype: :class:`injector.Injector`
        """
        ...
    


def process_dict(d: dict[Any, Any], injector: Injector) -> None:
    ...

def process_list(l: list[Any], injector: Injector) -> None:
    ...

class FlaskModule(Module):
    def __init__(self, app: flask.Flask, request_scope_class: type = ...) -> None:
        ...
    
    def configure(self, binder: Binder) -> None:
        ...
    





from connexion import FlaskApp
from connexion.middleware import MiddlewarePosition
from injector import Module, inject
from Ligare.web.config import Config
from starlette.middleware.cors import CORSMiddleware


class CORSMiddlewareModule(Module):
    @inject
    def register_middleware(self, app: FlaskApp, config: Config):
        cors_config = config.web.security.cors

        app.add_middleware(
            CORSMiddleware,
            position=MiddlewarePosition.BEFORE_EXCEPTION,
            allow_origins=cors_config.origins or [],
            allow_credentials=cors_config.allow_credentials,
            allow_methods=cors_config.allow_methods,
            allow_headers=cors_config.allow_headers,
        )



"""
This type stub file was generated by pyright.
"""

_signals = ...
user_logged_in = ...
user_logged_out = ...
user_loaded_from_cookie = ...
_user_loaded_from_header = ...
user_loaded_from_request = ...
user_login_confirmed = ...
user_unauthorized = ...
user_needs_refresh = ...
user_accessed = ...
session_protected = ...
def __getattr__(name): # -> NamedSignal:
    ...




import importlib
import logging
import sys
from collections import defaultdict
from contextlib import _GeneratorContextManager  # pyright: ignore[reportPrivateUsage]
from contextlib import ExitStack
from dataclasses import dataclass
from types import ModuleType
from typing import (
    Any,
    Callable,
    Generator,
    Generic,
    NamedTuple,
    Protocol,
    Sequence,
    TypeVar,
    cast,
)
from unittest.mock import AsyncMock, MagicMock, NonCallableMagicMock

import json_logging
import pytest
from _pytest.fixtures import SubRequest
from connexion import FlaskApp
from flask import Flask, Request, Response, session
from flask.ctx import RequestContext
from flask.sessions import SecureCookieSession
from flask.testing import FlaskClient
from flask_injector import FlaskInjector
from injector import Module
from Ligare.database.migrations.alembic.env import set_up_database
from Ligare.identity.config import SAML2Config, SSOConfig
from Ligare.platform.dependency_injection import UserLoaderModule
from Ligare.platform.identity import Role, User
from Ligare.platform.identity.user_loader import TRole, UserId, UserMixin
from Ligare.programming.collections.dict import NestedDict
from Ligare.programming.config import AbstractConfig, ConfigBuilder
from Ligare.programming.str import get_random_str
from Ligare.web.application import (
    ApplicationBuilder,
    CreateAppResult,
    FlaskAppResult,
    OpenAPIAppResult,
    T_app,
)
from Ligare.web.config import (
    Config,
    FlaskConfig,
    FlaskOpenApiConfig,
    FlaskSessionConfig,
    FlaskSessionCookieConfig,
)
from Ligare.web.encryption import encrypt_flask_cookie
from Ligare.web.middleware.sso import SAML2MiddlewareModule
from mock import MagicMock
from pytest import FixtureRequest
from pytest_mock import MockerFixture
from pytest_mock.plugin import MockType
from sqlalchemy.engine import Connection
from sqlalchemy.orm import Session
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.testclient import TestClient
from typing_extensions import Self
from werkzeug.local import LocalProxy

TFlaskClient = FlaskClient | TestClient
T_flask_client = TypeVar("T_flask_client", bound=TFlaskClient)


@dataclass
class ClientInjector(Generic[T_flask_client]):
    """
    Flask test client and IoC container for the application
    created by PyTest fixtures.
    """

    client: T_flask_client
    injector: FlaskInjector


class AppGetter(Protocol[T_app]):
    """
    A callable that instantiates a Flask application
    and returns the application with its IoC container.
    """

    def __call__(
        self,
    ) -> CreateAppResult[T_app]: ...


TAppInitHook = (
    Callable[[list[type[AbstractConfig]], list[Module | type[Module]]], None] | None
)
TClientInitHook = Callable[[CreateAppResult[T_app]], None] | None


class ClientInjectorConfigurable(Protocol[T_app, T_flask_client]):
    """
    Get a Flask test client using the specified application configuration.

    Args
    ------
        config: `Config` The custom application configuration used to instantiate the Flask app.

    Returns
    ------
        `FlaskClientInjector[T_flask_client]`
    """

    def __call__(
        self,
        config: Config,
        client_init_hook: TClientInitHook[T_app] | None = None,
        app_init_hook: TAppInitHook | None = None,
    ) -> Generator[ClientInjector[T_flask_client], Any, None]: ...


FlaskClientInjectorConfigurable = ClientInjectorConfigurable[Flask, FlaskClient]
OpenAPIClientInjectorConfigurable = ClientInjectorConfigurable[FlaskApp, TestClient]
FlaskClientInjector = ClientInjector[FlaskClient]
OpenAPIClientInjector = ClientInjector[TestClient]
OpenAPIClientInjectorWithDatabase = tuple[OpenAPIClientInjector, Connection]
OpenAPIClientInitHook = TClientInitHook[FlaskApp]
OpenAPIAppInitHook = TAppInitHook


class RequestConfigurable(Protocol):
    """
    Get a Flask request context, creating a Flask test client
    that uses the specified application configuration and,
    optionally, specific request context arguments.

    This creates the Flask test client using the `ClientConfigurable` fixture.

    Args
    ------
        config: `Config` The custom application configuration used to instantiate the Flask app.
        request_context_args: `dict[Any, Any] | None = None` The optional request context arguments
            set up in the request context. These may contain, for example, request headers,
            authentication credentials, etc.

    Returns
    ------
        `RequestContext`
    """

    def __call__(
        self, config: Config, request_context_args: dict[Any, Any] | None = None
    ) -> RequestContext: ...


class TestSessionMiddleware:
    _app: Flask

    def __init__(self, app: Flask) -> None:  # pyright: ignore[reportMissingSuperCall]
        self._app = app

    def __call__(
        self, environ: dict[Any, Any], start_response: Callable[[Request], Response]
    ) -> Any:
        from flask import request

        session["_id"] = get_random_str()

        return start_response(request)


class MockController(NamedTuple):
    begin: Callable[[], None]
    end: Callable[[], None]


class OpenAPIMockController(MockController):
    """
    A fixture used to mock some internals of Connexion to enable
    GET requests in tests.

    An OpenAPI application should be created, and any bindings managed,
    before calling `openapi_mock.begin()`. Once `begin()` is called,
    changes to the application may result in undefined behavior.

    `openapi_mock.end()` is called after the test finishes, and may also
    be explicitly called in the test.

    Args
    ------
        `Callable[[], Response]` A method used as the request handler for GET requests set to `/`.
        This parameter is set through parametrization:

        ```python
        @pytest.mark.parametrize("openapi_mock", lambda: {}, indirect=["openapi_mock"])
        def test_my_test(openapi_mock: OpenAPIMockController):
            ...
        ```

    Returns
    ------
        `RequestContext`
    """


class CreateApp(Generic[T_app]):
    app_name: str = "test_app"
    auto_mock_dependencies: bool = True
    default_app_getter: AppGetter[T_app] | None = None
    use_inmemory_database: bool = True

    _automatic_mocks: dict[str, MagicMock] = defaultdict()

    def _basic_config(self) -> Config:
        config = (
            ConfigBuilder[Config]()
            .with_root_config(Config)
            .with_configs([SSOConfig])
            .build()
        )()
        config.flask = FlaskConfig(
            app_name=self.app_name,
            env="Testing",
            session=FlaskSessionConfig(
                cookie=FlaskSessionCookieConfig(secret_key=get_random_str())
            ),
        )
        return config

    def _openapi_config(self) -> Config:
        config = self._basic_config()
        cast(FlaskConfig, config.flask).openapi = FlaskOpenApiConfig(
            spec_path="openapi.yaml", use_swagger=False
        )
        return config

    @pytest.fixture()
    def basic_config(self):
        return self._basic_config()

    @pytest.fixture()
    def openapi_config(self):
        return self._openapi_config()

    _MOCK_USER_USERNAME = "test user"

    def get_authenticated_request_context(
        self,
        app: ClientInjector[T_flask_client],
        user: type[UserMixin[TRole]],
        mocker: MockerFixture,
        roles: Sequence[TRole] | None = None,
    ):
        _ = self.mock_user(user, mocker, roles)

        with app.injector.app.test_request_context() as request_context:
            session["_id"] = get_random_str()
            session["authenticated"] = True
            session["username"] = CreateApp._MOCK_USER_USERNAME

            app_config = cast(dict[str, Any], app.injector.app.config)

            if isinstance(app.client, TestClient):
                session_cookie = encrypt_flask_cookie(
                    # TODO use app.client.injector?
                    app_config["SECRET_KEY"],
                    {
                        "_id": session["_id"],
                        "authenticated": True,
                        "username": CreateApp._MOCK_USER_USERNAME,
                    },
                )

                app.client.cookies.set(
                    app_config["SESSION_COOKIE_NAME"], session_cookie
                )
            else:
                # TODO move the session_transaction stuff from the Flask test client stuff here.
                raise NotImplementedError(
                    f"Authenticated request context for `{FlaskClient.__name__}` not implemented."
                )

            return request_context

    def mock_user(
        self,
        user: type[UserMixin[TRole]],
        mocker: MockerFixture,
        roles: Sequence[TRole] | None = None,
    ) -> MagicMock | AsyncMock | NonCallableMagicMock:
        if roles is None:
            roles = cast(Sequence[TRole], ())

        def get_mock_user(proxy: LocalProxy[UserMixin[TRole]] | None = None):
            if proxy is not None:
                return proxy
            user_id = UserId(1, CreateApp._MOCK_USER_USERNAME)
            # FIXME this needs to handle user roles
            return user(user_id, roles)

        return mocker.patch("flask_login.utils._get_user", side_effect=get_mock_user)

    def _client_configurable(
        self,
        mocker: MockerFixture,
        app_getter: Callable[
            [Config, MockerFixture, TAppInitHook | None],
            Generator[CreateAppResult[T_app], Any, None],
        ],
        client_getter: Callable[
            [AppGetter[T_app]],
            Generator[ClientInjector[T_flask_client], Any, None],
        ],
    ) -> ClientInjectorConfigurable[T_app, T_flask_client]:
        def _client_getter(
            config: Config,
            client_init_hook: TClientInitHook[T_app] | None = None,
            app_init_hook: TAppInitHook | None = None,
        ) -> Generator[ClientInjector[T_flask_client], Any, None]:
            application_result = next(app_getter(config, mocker, app_init_hook))

            if client_init_hook is not None:
                client_init_hook(application_result)

            app_result = next(client_getter(lambda: application_result))
            with app_result.injector.app.app_context():
                yield app_result

        return _client_getter

    # https://stackoverflow.com/a/55079736
    # creates a fixture on this class called `setup_method_fixture`
    # then tells pytest to use it for every test in the class
    @pytest.fixture(autouse=True)
    def setup_method_fixture(self, mocker: MockerFixture):
        setup_artifacts = self._pre_test_setup(mocker)
        yield
        self._post_test_teardown(setup_artifacts)

    def _pre_test_setup(self, mocker: MockerFixture):
        # the pytest log formatters need to be restored
        # in the event json_logging changes them, otherwise
        # some tests may fail
        log_formatters = [handler.formatter for handler in logging.getLogger().handlers]

        self._automatic_mocks = {}

        mock_targets: list[tuple[str] | tuple[str, Any]] = [
            ("io.open",),
            ("toml.decoder.loads", {}),
            ("Ligare.web.application._import_blueprint_modules", []),
            ("Ligare.web.application._get_program_dir", "."),
            ("Ligare.web.application._get_exec_dir", ".."),
        ]

        if self.auto_mock_dependencies:
            for mock_target in mock_targets:
                if len(mock_target) == 1:
                    (target_name,) = mock_target
                    mock = mocker.patch(target_name)
                else:
                    (target_name, return_value) = mock_target
                    mock = mocker.patch(target_name)
                    mock.return_value = return_value

                self._automatic_mocks[target_name] = mock

        return log_formatters

    def _post_test_teardown(self, log_formatters: list[logging.Formatter | None]):
        self._automatic_mocks = {}

        for i, handler in enumerate(logging.getLogger().handlers):
            # this assumes handlers are in the same order
            # so is prone to breakage
            handler.formatter = log_formatters[i]

        # json_logging relies on global, so they must be reset between each test
        _ = importlib.reload(json_logging)


class CreateFlaskApp(CreateApp[Flask]):
    def __get_basic_flask_app(
        self,
        config: Config,
        mocker: MockerFixture,
        app_init_hook: TAppInitHook | None = None,
    ) -> Generator[FlaskAppResult, Any, None]:
        # prevents the creation of a Connexion application
        if config.flask is not None:
            config.flask.openapi = None
            config.sso = SSOConfig(  # pyright: ignore[reportAttributeAccessIssue]
                protocol="SAML2",
                settings=SAML2Config(relay_state="", metadata_url="", metadata=""),
            )

        _ = mocker.patch("Ligare.web.application.load_config", return_value=config)
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=config)),
        )

        application_configs: list[type[AbstractConfig]] | None = []
        application_modules: list[Module | type[Module]] | None = []

        if app_init_hook is not None:
            app_init_hook(application_configs, application_modules)

        application_modules.append(SAML2MiddlewareModule)

        logging.basicConfig(force=True)

        application_builder = (
            ApplicationBuilder[Flask]()
            .with_modules(application_modules)
            .use_configuration(
                lambda config_builder: config_builder.enable_ssm(True)
                .with_config_filename("config.toml")
                .with_root_config_type(Config)
                .with_config_types(application_configs)
            )
        )
        app = application_builder.build()
        yield app

    @pytest.fixture()
    def _get_basic_flask_app(
        self, basic_config: Config, mocker: MockerFixture
    ) -> FlaskAppResult:
        return next(self.__get_basic_flask_app(basic_config, mocker))

    def _flask_client(
        self, flask_app_getter: AppGetter[Flask]
    ) -> Generator[FlaskClientInjector, Any, None]:
        with ExitStack() as stack:
            result = flask_app_getter()

            if not isinstance(result, CreateAppResult) or not isinstance(  # pyright: ignore[reportUnnecessaryIsInstance]
                result.app_injector.app, Flask
            ):
                raise Exception(
                    f"""This fixture created a `{type(result)}.{type(result.app_injector.app) if getattr(result, "app", None) else "None"}` application, but is only meant for `{Flask}`.
Ensure either that [openapi] is not set in the [flask] config, or use the `openapi_client` fixture."""
                )

            app = result.app_injector.app
            client = stack.enter_context(app.test_client())

            app = client.application

            flask_session_ctx_manager = cast(
                _GeneratorContextManager[SecureCookieSession],
                client.session_transaction(),
            )
            flask_session = stack.enter_context(flask_session_ctx_manager)
            flask_session["_id"] = get_random_str()
            client.set_cookie(
                cast(str, app.config["SESSION_COOKIE_NAME"]),
                encrypt_flask_cookie(
                    cast(str, app.config["SECRET_KEY"]), flask_session
                ),
                domain="localhost",
                # fmt: off
                max_age=app.config["PERMANENT_SESSION_LIFETIME"]
                if app.config["PERMANENT_SESSION"]
                else None,
                # fmt: on
            )

            yield ClientInjector(client, result.app_injector.flask_injector)

    @pytest.fixture()
    def flask_client(self, _get_basic_flask_app: FlaskAppResult) -> FlaskClientInjector:
        return next(self._flask_client(lambda: _get_basic_flask_app))

    @pytest.fixture()
    def flask_client_configurable(
        self, mocker: MockerFixture
    ) -> FlaskClientInjectorConfigurable:
        return self._client_configurable(
            mocker, self.__get_basic_flask_app, self._flask_client
        )

    def _flask_request(
        self,
        flask_client: FlaskClient,
        request_context_args: dict[Any, Any] | None = None,
    ) -> Generator[RequestContext, Any, None]:
        with flask_client.application.test_request_context(
            **(request_context_args or {})
        ) as request_context:
            yield request_context

    @pytest.fixture()
    def flask_request(
        self, flask_client: FlaskClient
    ) -> Generator[RequestContext, Any, None]:
        with next(self._flask_request(flask_client)) as request_context:
            yield request_context

    @pytest.fixture()
    def flask_request_configurable(
        self,
        flask_client_configurable: FlaskClientInjectorConfigurable,
    ) -> RequestConfigurable:
        def _flask_request_getter(
            config: Config, request_context_args: dict[Any, Any] | None = None
        ):
            flask_client = next(flask_client_configurable(config))
            request_context = next(
                self._flask_request(flask_client.client, request_context_args)
            )
            return request_context

        return _flask_request_getter


class CreateOpenAPIApp(CreateApp[FlaskApp]):
    def _get_real_openapi_app(
        self,
        config: Config,
        mocker: MockerFixture,
        app_init_hook: TAppInitHook | None = None,
    ) -> Generator[OpenAPIAppResult, Any, None]:
        # prevents the creation of a Connexion application
        if config.flask is None or config.flask.openapi is None:
            raise Exception(
                "[openapi] not set in config. Cannot create OpenAPI test client."
            )

        _ = mocker.patch("Ligare.web.application.load_config", return_value=config)
        _ = mocker.patch(
            "Ligare.web.application.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=config)),
        )

        _application_configs: list[type[AbstractConfig]] = []
        _application_modules = cast(
            list[Module | type[Module]],
            [
                SAML2MiddlewareModule,
                UserLoaderModule(
                    loader=User,  # pyright: ignore[reportArgumentType]
                    roles=Role,  # pyright: ignore[reportArgumentType]
                    user_table=MagicMock(),  # pyright: ignore[reportArgumentType]
                    role_table=MagicMock(),  # pyright: ignore[reportArgumentType]
                    bases=[],
                ),
            ],
        )

        if app_init_hook is not None:
            app_init_hook(_application_configs, _application_modules)

        application_builder = (
            ApplicationBuilder[FlaskApp]()
            .with_modules(_application_modules)
            .use_configuration(
                lambda config_builder: config_builder.enable_ssm(True)
                .with_config_filename("config.toml")
                .with_root_config_type(Config)
                .with_config_types(_application_configs)
            )
        )
        app = application_builder.build()
        yield app

    @pytest.fixture()
    def _get_openapi_app(
        self,
        openapi_config: Config,
        mocker: MockerFixture,
        openapi_mock_controller: OpenAPIMockController,
    ) -> OpenAPIAppResult:
        _ = mocker.patch("Ligare.web.application.json_logging")
        openapi_mock_controller.begin()
        return next(self._get_real_openapi_app(openapi_config, mocker))

    def _openapi_client(
        self, flask_app_getter: AppGetter[FlaskApp]
    ) -> Generator[OpenAPIClientInjector, Any, None]:
        with ExitStack() as stack:
            result = flask_app_getter()

            if not isinstance(result, CreateAppResult) or not isinstance(  # pyright: ignore[reportUnnecessaryIsInstance]
                result.app_injector.app, FlaskApp
            ):
                raise Exception(
                    f"""This fixture created a `{type(result)}.{type(result.app_injector.app) if getattr(result, "app", None) else "None"}` application, but is only meant for `{FlaskApp}`.
Ensure either that [openapi] is set in the [flask] config, or use the `flask_client` fixture."""
                )
            config = result.app_injector.flask_injector.injector.get(Config)
            host = config.flask.host if config.flask else "localhost"
            port = config.flask.port if config.flask else "5000"
            result.app_injector.app.add_middleware(
                TrustedHostMiddleware,
                allowed_hosts=[host, f"{host}:{port}"],
            )

            client: TestClient = stack.enter_context(
                result.app_injector.app.test_client()
            )

            app_config = cast(dict[str, Any], result.app_injector.app.app.config)

            _ = client.headers.setdefault("Host", app_config["SERVER_NAME"])

            yield ClientInjector(client, result.app_injector.flask_injector)

    # TODO this needs to support plain old Flask at some point
    def get_app(self, flask_app_getter: AppGetter[FlaskApp]):
        return next(self._openapi_client(flask_app_getter))

    @pytest.fixture()
    def openapi_client_with_database(
        self, request: SubRequest
    ) -> Generator[tuple[OpenAPIClientInjector, Connection], Any, None]:
        if self.use_inmemory_database:
            _ = request.getfixturevalue("use_inmemory_database")

        openapi_client = request.getfixturevalue("openapi_client")

        with openapi_client.injector.injector.get(Session) as session:
            # Get a connection to share so Alembic does not wipe out the in-memory database
            # when using SQLite in-memory connections
            if session.bind is None:
                raise Exception(
                    "SQLAlchemy Session is not bound to an engine. This is not supported."
                )

            with set_up_database(session.bind.engine) as connection:
                yield (openapi_client, connection)

    @pytest.fixture()
    def openapi_client(
        self,
        request: SubRequest,
    ) -> OpenAPIClientInjector:
        # Call `default_app_getter` as a class method because:
        # 1. A method assigned as a value to a class member becomes a class
        #    member itself which means `self` is implicitly passed to the method
        # 2. `AppGetter[T_app]` does not take parameters, and the implicit `self`
        #    causes a runtime call error.
        # 3. The only way to avoid the class member behavior is to avoid
        #    assigning the class member value (during declaration), instead
        #    assigning it as an object value (during instantiation or later).
        # 4. Pytest does not load tests in a class that has an __init__ method
        #    so we cannot use a ctor to assign the value
        #
        # also:
        # assertions in self._openapi_client check the actual type,
        # so this cast is just to appease pyright.
        # TODO support this for Flask as well
        default_app_getter = cast(Self, self.__class__).default_app_getter

        get_openapi_app: CreateAppResult[FlaskApp]
        if default_app_getter is None:
            get_openapi_app = request.getfixturevalue("_get_openapi_app")
        else:
            get_openapi_app = default_app_getter()

        return next(self._openapi_client(lambda: get_openapi_app))

    @pytest.fixture()
    def openapi_client_configurable(
        self, mocker: MockerFixture
    ) -> OpenAPIClientInjectorConfigurable:
        # FIXME some day json_logging needs to be fixed
        _ = mocker.patch("Ligare.web.application.json_logging")
        return self._client_configurable(
            mocker, self._get_real_openapi_app, self._openapi_client
        )

    def _openapi_request(
        self,
        openapi_client: TestClient,
        request_context_args: dict[Any, Any] | None = None,
    ) -> Generator[RequestContext, Any, None]:
        with cast(FlaskApp, openapi_client.app).app.test_request_context(
            **(request_context_args or {})
        ) as request_context:
            yield request_context

    @pytest.fixture()
    def openapi_request(
        self, openapi_client: TestClient
    ) -> Generator[RequestContext, Any, None]:
        with next(self._openapi_request(openapi_client)) as request_context:
            yield request_context

    @pytest.fixture()
    def openapi_request_configurable(
        self,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
    ) -> RequestConfigurable:
        def _flask_request_getter(
            config: Config, request_context_args: dict[Any, Any] | None = None
        ):
            flask_client = next(openapi_client_configurable(config))
            request_context = next(
                self._openapi_request(flask_client.client, request_context_args)
            )
            return request_context

        return _flask_request_getter

    # this is the YAML-parsed dictionary from this OpenAPI spec
    # openapi: 3.0.3
    # servers:
    # - url: http://testserver/
    #    description: Test Application
    # info:
    # title: "Test Application"
    # version: 3.0.3
    # paths:
    # /:
    #    get:
    #    description: "Check whether the application is running."
    #    operationId: "root.get"
    #    parameters: []
    #    responses:
    #        "200":
    #        content:
    #            application/json:
    #            schema:
    #                type: string
    #        description: "Application is running correctly."
    #    summary: "A simple method that returns 200 as long as the application is running."
    _openapi_spec: NestedDict[str, Any] = {
        "info": {"title": "Test Application", "version": "3.0.3"},
        "openapi": "3.0.3",
        "paths": {
            "/": {
                "get": {
                    "description": "Check whether the application is running.",
                    "operationId": "root.get",
                    "parameters": [],
                    "responses": {
                        "200": {
                            "content": {
                                "application/json": {"schema": {"type": "string"}}
                            },
                            "description": "Application is running correctly",
                        }
                    },
                    "summary": "A simple method that returns 200 as long as the application is running.",
                }
            }
        },
        "servers": [{"description": "Test Application", "url": "http://testserver/"}],
    }

    @pytest.fixture()
    def openapi_mock_controller(self, request: FixtureRequest, mocker: MockerFixture):
        fixture_name = CreateOpenAPIApp.openapi_mock_controller.__name__
        parameter_error_msg = f"The first parameter to the `{fixture_name}` fixture must be a `Callable[[], Response]`. Review the documentation for `OpenAPIMockController`."

        get_handler: Callable[[], Response]
        # TODO probably log when this happens
        if not getattr(request, "param", None):
            get_handler = lambda: Response()
        else:
            get_handler = request.param

        if not callable(get_handler):
            raise Exception(parameter_error_msg)

        openapi_spec_module_name = "root"
        openapi_spec_root_module = ModuleType(openapi_spec_module_name)

        setattr(openapi_spec_root_module, "get", lambda: get_handler())
        sys.modules[openapi_spec_module_name] = openapi_spec_root_module
        openapi_spec_import_module = MagicMock(return_value=openapi_spec_root_module)

        # don't create the app, just prep for
        # the app to load all these mocks instead
        importlib_mock: MockType | None = None
        spec_loader_mock: MockType | None = None

        def begin():
            nonlocal importlib_mock, spec_loader_mock

            if importlib_mock is None:
                importlib_mock = mocker.patch(
                    "connexion.utils.importlib",
                    spec=importlib,
                    import_module=openapi_spec_import_module,
                )

            if spec_loader_mock is None:
                spec_loader_mock = mocker.patch(
                    "connexion.spec.Specification._load_spec_from_file",
                    return_value=CreateOpenAPIApp._openapi_spec,
                )

        def end():
            nonlocal importlib_mock, spec_loader_mock

            if importlib_mock is not None:
                mocker.stop(importlib_mock)

            if spec_loader_mock is not None:
                mocker.stop(spec_loader_mock)

        mock_controller = MockController(begin=begin, end=end)

        try:
            # TODO can this be a context manager instead of requiring
            # the explicit begin() call?
            yield mock_controller
        finally:
            mock_controller.end()
            _ = sys.modules.pop(openapi_spec_module_name)



"""
This type stub file was generated by pyright.
"""

import json
import typing as t

"""
This module centralizes all functionality related to json encoding and decoding in Connexion.
"""
def wrap_default(default_fn: t.Callable) -> t.Callable:
    """The Connexion defaults for JSON encoding. Handles extra types compared to the
    built-in :class:`json.JSONEncoder`.

    -   :class:`datetime.datetime` and :class:`datetime.date` are
        serialized to :rfc:`822` strings. This is the same as the HTTP
        date format.
    -   :class:`decimal.Decimal` is serialized to a float.
    -   :class:`uuid.UUID` is serialized to a string.
    """
    ...

class JSONEncoder(json.JSONEncoder):
    """The default Connexion JSON encoder. Handles extra types compared to the
    built-in :class:`json.JSONEncoder`.

    -   :class:`datetime.datetime` and :class:`datetime.date` are
        serialized to :rfc:`822` strings. This is the same as the HTTP
        date format.
    -   :class:`uuid.UUID` is serialized to a string.
    """
    @wrap_default
    def default(self, o): # -> Any:
        ...
    


class Jsonifier:
    """
    Central point to serialize and deserialize to/from JSon in Connexion.
    """
    def __init__(self, json_=..., **kwargs) -> None:
        """
        :param json_: json library to use. Must have loads() and dumps() method  # NOQA
        :param kwargs: default arguments to pass to json.dumps()
        """
        ...
    
    def dumps(self, data, **kwargs): # -> str:
        """Central point where JSON serialization happens inside
        Connexion.
        """
        ...
    
    def loads(self, data): # -> Any | str | None:
        """Central point where JSON deserialization happens inside
        Connexion.
        """
        ...
    


